import contextlib
import sys
from typing import TYPE_CHECKING, Iterator

import qik.console

if TYPE_CHECKING:
    import qik.ctx as qik_ctx
else:
    import qik.lazy

    qik_ctx = qik.lazy.module("qik.ctx")


class Error(Exception):
    code = "general"


class RunnerError(Error):
    """Runner errors result in the whole runner exiting."""

    code = "runner"


class ConfigNotFound(RunnerError):
    code = "conf0"


class ConfigParse(RunnerError):
    code = "conf1"


class ModuleNotFound(RunnerError):
    code = "conf2"


class PluginNotFound(RunnerError):
    code = "conf3"


class ModuleOrPluginNotFound(RunnerError):
    code = "conf4"


class CommandNotFound(RunnerError):
    code = "conf5"


class PluginImport(RunnerError):
    code = "conf6"


class GraphCycle(RunnerError):
    code = "conf7"


class SpaceNotFound(RunnerError):
    code = "conf8"


class UnconfiguredCache(RunnerError):
    code = "conf9"


class InvalidCacheType(RunnerError):
    code = "conf10"


class InvalidDepType(RunnerError):
    code = "conf11"


class InvalidVenvType(RunnerError):
    code = "conf12"


class CircularVenv(RunnerError):
    code = "conf13"


class CircularConstraint(RunnerError):
    code = "conf14"


class CircularFence(RunnerError):
    code = "conf14"


class CtxProfileNotFound(RunnerError):
    code = "ctx0"


class EnvCast(RunnerError):
    code = "ctx1"


class CtxValueNotFound(RunnerError):
    code = "ctx2"


class UnconfiguredCtx(RunnerError):
    code = "ctx3"


class InvalidCtxNamespace(RunnerError):
    code = "ctx4"


class ArgNotSupplied(RunnerError):
    code = "args0"


class RunnableError(Error):
    """Runnable errors result in an individual runnable erroring."""

    code = "runnable"


class LockFileNotFound(RunnableError):
    code = "space0"


class VenvNotFound(RunnableError):
    code = "space1"


class DotEnvNotFound(RunnableError):
    code = "space2"


class ModuleDistributionNotFound(RunnableError):
    code = "pydist0"


class DistributionNotFound(RunnableError):
    code = "pydist1"


def fmt_msg(exc: Exception, prefix: str = "") -> str:
    err_kwargs = (
        {} if isinstance(exc, RunnableError) else {"emoji": "broken_heart", "color": "red"}
    )
    if isinstance(exc, Error):
        return qik.console.fmt_msg(
            f"{prefix}{exc.args[0]} [reset][dim]See https://qik.build/en/stable/errors/#{exc.code}[/dim]",
            **err_kwargs,  # type: ignore
        )
    else:
        return qik.console.fmt_msg("An unexpected error happened", **err_kwargs)  # type: ignore


def print(exc: Exception, prefix: str = "") -> None:
    msg = fmt_msg(exc, prefix=prefix)
    qik.console.get().print(msg)
    if not isinstance(exc, Error) or qik_ctx.by_namespace("qik").verbosity >= 3:
        qik.console.print_exception()


@contextlib.contextmanager
def catch_and_exit() -> Iterator[None]:
    try:
        yield
    except Error as e:
        print(e)
        sys.exit(1)



from __future__ import annotations

import qik.conf


class S3Conf(qik.conf.Cache, frozen=True, tag="s3"):
    bucket: str
    prefix: str = ""
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    aws_session_token: str | None = None
    region_name: str | None = None
    endpoint_url: str | None = None


qik.conf.register_type(S3Conf, "qik.s3.cache.factory")



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING

import qik.cmd
import qik.conf
import qik.dep
import qik.errors
import qik.func
import qik.space
import qik.unset
import qik.uv.conf
import qik.uv.utils
import qik.venv

if TYPE_CHECKING:
    from qik.uv.qikplugin import UVVenvConf


def _resolve_constraint(
    constraint: str | qik.conf.SpaceLocator | qik.unset.UnsetType | None,
) -> str | qik.unset.UnsetType | None:
    if isinstance(constraint, str):
        return constraint
    elif isinstance(constraint, qik.conf.SpaceLocator):
        venv = qik.space.load(constraint.name).venv
        return (
            _resolve_constraint(venv.conf.constraint)
            if isinstance(venv, UVVenv)
            else qik.unset.UNSET
        )
    else:
        return constraint


class UVVenv(qik.venv.Venv, frozen=True, dict=True):
    conf: UVVenvConf

    @qik.func.cached_property
    def python(self) -> str | None:
        return qik.unset.coalesce(self.conf.python, qik.uv.conf.get().python, default=None)

    @qik.func.cached_property
    def constraint(self) -> str | None:
        try:
            return qik.unset.coalesce(
                _resolve_constraint(self.conf.constraint),
                _resolve_constraint(qik.uv.conf.get().constraint),
                default=None,
            )
        except RecursionError as e:
            raise qik.errors.CircularConstraint("Circular constraint detected.") from e

    @qik.func.cached_property
    def default_lock(self) -> str:
        uv_conf = qik.uv.conf.get()
        root = (
            qik.conf.pub_work_dir(rel=True)
            if uv_conf.resolved_cache == "repo"
            else qik.conf.priv_work_dir(rel=True)
        )
        return str(
            root
            / "artifacts"
            / qik.uv.utils.lock_cmd_name()
            / f"requirements-{self.name}-lock.txt"
        )

    @qik.func.cached_property
    def lock(self) -> str:
        super_lock = super().lock
        return self.default_lock if not super_lock else super_lock

    @qik.func.cached_property
    def environ(self) -> dict[str, str]:  # type: ignore
        return os.environ | {
            "VIRTUAL_ENV": str(self.dir),
            "PATH": f"{self.dir}/bin:{os.environ['PATH']}",
        }

    @qik.func.cached_property
    def dir(self) -> pathlib.Path:  # type: ignore
        return qik.conf.priv_work_dir() / "venv" / self.name

    @qik.func.cached_property
    def site_packages_dir(self) -> pathlib.Path:  # type: ignore
        for path in pathlib.Path(self.dir).glob("lib/python*/site-packages"):
            return path

        # TODO: Turn this into a qik runtime error
        raise AssertionError(
            f'Could not find site packages dir of venv "{self.name}" at "{self.dir}"'
        )

    @qik.func.cached_property
    def runnable_deps(self) -> dict[str, qik.dep.Runnable]:
        return {
            runnable.name: qik.dep.Runnable(name=runnable.name, obj=runnable, strict=True)
            for runnable in qik.cmd.load(
                qik.uv.utils.install_cmd_name(), space=self.name
            ).runnables.values()
        }


def factory(name: str, conf: UVVenvConf) -> UVVenv:
    return UVVenv(name=name, conf=conf)



from __future__ import annotations

from typing import TYPE_CHECKING

import qik.conf
import qik.func
import qik.pygraph.conf

if TYPE_CHECKING:
    import pathlib


@qik.func.cache
def build_cmd_name() -> str:
    graph_plugin_name = qik.conf.plugin_locator("qik.pygraph", by_pyimport=True)[0]
    return f"{graph_plugin_name}.build"


@qik.func.cache
def lock_cmd_name() -> str:
    graph_plugin_name = qik.conf.plugin_locator("qik.pygraph", by_pyimport=True)[0]
    return f"{graph_plugin_name}.lock"


@qik.func.cache
def check_cmd_name() -> str:
    graph_plugin_name = qik.conf.plugin_locator("qik.pygraph", by_pyimport=True)[0]
    return f"{graph_plugin_name}.check"


@qik.func.cache
def graph_path(rel: bool = True) -> pathlib.Path:
    pygraph_conf = qik.pygraph.conf.get()
    root = (
        qik.conf.pub_work_dir(rel=rel)
        if pygraph_conf.resolved_build_cache == "repo"
        else qik.conf.priv_work_dir(rel=rel)
    )
    return root / "artifacts" / build_cmd_name() / "graph.json"


@qik.func.cache
def lock_path(pyimport: str, space: str | None = None, rel: bool = True) -> pathlib.Path:
    pygraph_conf = qik.pygraph.conf.get()
    root = (
        qik.conf.pub_work_dir(rel=rel)
        if pygraph_conf.resolved_lock_cache == "repo"
        else qik.conf.priv_work_dir(rel=rel)
    )
    file_name = f"lock.{pyimport}.{space}.json" if space else f"lock.{pyimport}.json"
    return root / "artifacts" / lock_cmd_name() / file_name



from typing import ClassVar

import qik.conf
import qik.ctx
import qik.func
import qik.unset


class UVVenvConf(qik.conf.Venv, frozen=True, tag="uv"):
    python: str | qik.unset.UnsetType = qik.unset.UNSET
    constraint: str | qik.conf.SpaceLocator | qik.unset.UnsetType = qik.unset.UNSET
    install_cmd: ClassVar[str] = "uv.install"


class UVPluginConf(qik.conf.PluginConf, frozen=True, dict=True, tag="qik.uv"):
    cache: str | qik.unset.UnsetType = qik.unset.UNSET
    python: str | None = None
    index_url: str | None = None
    extra_index_url: str | None = None
    constraint: str | qik.conf.SpaceLocator | None = None

    @qik.func.cached_property
    def resolved_index_url(self) -> str | None:
        if self.index_url:
            return qik.ctx.format(self.index_url)

    @qik.func.cached_property
    def resolved_extra_index_url(self) -> str | None:
        if self.extra_index_url:
            return qik.ctx.format(self.extra_index_url)

    @qik.func.cached_property
    def resolved_cache(self) -> str:
        return qik.ctx.format(
            qik.unset.coalesce(self.cache, qik.conf.project().defaults.cache, default="repo")
        )


qik.conf.register_type(UVVenvConf, "qik.uv.venv.factory")
qik.conf.register_conf(UVPluginConf)



from __future__ import annotations

from typing import TYPE_CHECKING

import qik.conf
import qik.func
import qik.uv.qikplugin

if TYPE_CHECKING:
    from qik.uv.qikplugin import UVPluginConf


@qik.func.cache
def get() -> UVPluginConf:
    """Get the uv config."""
    return qik.conf.plugin("qik.uv")  # type: ignore



from __future__ import annotations

import base64
import pathlib
import pkgutil
import threading
from typing import TYPE_CHECKING, Iterator

import msgspec

import qik.conf
import qik.ctx
import qik.errors
import qik.file
import qik.func
import qik.shell

if TYPE_CHECKING:
    from qik.runnable import Result, Runnable


# A global lock for git index manipulation operations
_LOCK = threading.Lock()


class Manifest(msgspec.Struct, frozen=True, omit_defaults=True):
    name: str
    hash: str
    code: int
    log: str | None = None
    artifacts: list[str] = []


class Entry(msgspec.Struct, frozen=True):
    manifest: Manifest
    log: str | None = None


def _walk_artifacts(runnable: Runnable) -> Iterator[pathlib.Path]:
    """Walk artifact globs of a runnable."""
    for artifact in set(runnable.artifacts):
        yield from qik.conf.root().glob(artifact)


def _artifact_name(path: str | pathlib.Path) -> str:
    return f"artifact-{base64.urlsafe_b64encode(str(path).encode()).decode()}"


class Cache:
    @property
    def type(self) -> str:
        return self.__class__.__name__.lower()

    def base_path(self, *, runnable: Runnable, hash: str) -> pathlib.Path:
        raise NotImplementedError

    def manifest_path(self, *, runnable: Runnable, hash: str) -> pathlib.Path:
        return self.base_path(runnable=runnable, hash=hash) / f"{runnable.slug}.json"

    def log_path(self, *, runnable: Runnable, hash: str) -> pathlib.Path:
        return self.base_path(runnable=runnable, hash=hash) / f"{runnable.slug}.out"

    def pre_get(self, *, runnable: Runnable, hash: str) -> None:
        pass

    def post_set(self, *, runnable: Runnable, hash: str, manifest: Manifest) -> None:
        pass

    def on_miss(self, *, runnable: Runnable, hash: str) -> None:
        raise NotImplementedError

    def restore_artifacts(self, *, runnable: Runnable, hash: str, artifacts: list[str]) -> None:
        pass

    def import_artifacts(self, *, runnable: Runnable, hash: str) -> list[str]:
        return []

    def get(self, runnable: Runnable, artifacts: bool = True) -> Entry | None:
        hash = runnable.hash()
        self.pre_get(runnable=runnable, hash=hash)

        base_path = self.base_path(runnable=runnable, hash=hash)
        manifest_path = self.manifest_path(runnable=runnable, hash=hash)

        def _get_entry() -> Entry:
            manifest = msgspec.json.decode(manifest_path.read_bytes(), type=Manifest)
            if manifest.hash != hash:
                raise FileNotFoundError("Manifest not found.")

            log = pathlib.Path(base_path / manifest.log).read_text() if manifest.log else None
            if artifacts:
                self.restore_artifacts(runnable=runnable, hash=hash, artifacts=manifest.artifacts)
            return Entry(manifest=manifest, log=log)

        try:
            return _get_entry()
        except FileNotFoundError:
            try:
                self.on_miss(runnable=runnable, hash=hash)
            except NotImplementedError:
                return None

        try:
            return _get_entry()
        except FileNotFoundError:
            return None

    def set(self, runnable: Runnable, result: Result) -> None:
        manifest_path = self.manifest_path(runnable=runnable, hash=result.hash)
        log_path = self.log_path(runnable=runnable, hash=result.hash)

        artifacts = self.import_artifacts(runnable=runnable, hash=result.hash)
        manifest = Manifest(
            name=runnable.name,
            code=result.code,
            hash=result.hash,
            log=log_path.name if result.log else None,
            artifacts=artifacts,
        )
        qik.file.write(manifest_path, msgspec.json.encode(manifest))
        if result.log:
            qik.file.write(log_path, result.log)

        self.post_set(runnable=runnable, hash=result.hash, manifest=manifest)


class Uncached(Cache):
    @property
    def type(self) -> str:
        return "none"

    def get(self, runnable: Runnable, artifacts: bool = True) -> Entry | None:
        pass

    def set(self, runnable: Runnable, result: Result) -> None:
        pass


@qik.func.cache
def _install_custom_merge_driver():
    """Install qik's custom git merge driver."""
    script_path = pathlib.Path(__file__).parent / "merge.sh"
    custom_merge_driver_install = f'git config merge.qik.driver "sh {script_path} %O %A %B"'
    qik.shell.exec(custom_merge_driver_install)


@qik.func.cache
def _add_cache_dir_to_git_attributes():
    """Add .qik to .gitattributes so that it appears differently in diff.

    https://docs.github.com/en/repositories/working-with-files/managing-files/customizing-how-changed-files-appear-on-github
    """
    git_root_dir = pathlib.Path(qik.shell.exec("git rev-parse --show-toplevel").stdout.strip())
    attrs_path = git_root_dir / ".gitattributes"
    ignore_glob = qik.conf.root().relative_to(git_root_dir) / ".qik/**/*"
    attrs_line = f"{ignore_glob} linguist-generated=true merge=qik\n"
    try:
        gitattributes = attrs_path.read_text()
        if attrs_line not in gitattributes:
            attrs_path.write_text(f"{attrs_line}{gitattributes}")
    except FileNotFoundError:
        attrs_path.write_text(attrs_line)
        qik.shell.exec(f"git add -N {attrs_path}")


class Repo(Cache):
    """A cache in the local git repository."""

    def base_path(self, *, runnable: Runnable, hash: str) -> pathlib.Path:
        return qik.conf.pub_work_dir() / "cache" / runnable.cmd

    def post_set(self, *, runnable: Runnable, hash: str, manifest: Manifest) -> None:
        git_add = [str(self.manifest_path(runnable=runnable, hash=hash))]

        if manifest.log:
            git_add.append(str(self.log_path(runnable=runnable, hash=hash)))

        if runnable.artifacts:
            git_add.extend(runnable.artifacts)

        git_add_files = " ".join(git_add)
        with _LOCK:
            qik.shell.exec(f"git add -N {git_add_files}")
            _add_cache_dir_to_git_attributes()
            _install_custom_merge_driver()


class Local(Cache):
    """A local cache in the ._qik directory."""

    def base_path(self, *, runnable: Runnable, hash: str) -> pathlib.Path:
        return qik.conf.priv_work_dir() / "cache" / runnable.cmd

    def restore_artifacts(self, *, runnable: Runnable, hash: str, artifacts: list[str]) -> None:
        base_path = self.base_path(runnable=runnable, hash=hash)
        for artifact in set(artifacts):
            name = _artifact_name(artifact)
            qik.file.copy(str(base_path / name), artifact)

    def import_artifacts(self, *, runnable: Runnable, hash: str) -> list[str]:
        base_path = self.base_path(runnable=runnable, hash=hash)
        artifacts = [str(path) for path in _walk_artifacts(runnable)]
        for artifact in artifacts:
            name = _artifact_name(artifact)
            qik.file.copy(artifact, str(base_path / name))

        return artifacts


@qik.func.cache
def load(name: str) -> Cache:
    proj = qik.conf.project()

    match name:
        case "repo":
            return Repo()
        case "local":
            return Local()
        case "none":
            return Uncached()
        case custom:
            if conf := proj.caches.get(custom):
                factory = qik.conf.get_type_factory(conf)
                return pkgutil.resolve_name(factory)(name, conf)
            else:
                raise qik.errors.UnconfiguredCache(f'Unconfigured cache - "{custom}"')



from typing import Any, TypeAlias, TypeVar

UnsetType: TypeAlias = frozenset
T = TypeVar("T")

UNSET = frozenset([None])


def coalesce(*args: Any, default: Any) -> Any:
    return next((arg for arg in args if not isinstance(arg, UnsetType | None)), default)



"""Core watching functionality.

Note that we have an integration test for this in test_integration.py::test_watch,
however coverage is not yet instrumented for it (hence pragma: no covers)
"""

from __future__ import annotations

import fnmatch
import pathlib
import re
import sys
import threading
from typing import TYPE_CHECKING

import qik.dep
import qik.func
import qik.space
import qik.unset

if TYPE_CHECKING:
    import watchdog.events as watchdog_events
    import watchdog.observers as watchdog_observers

    import qik.venv
    from qik.runner import Runner

    class QikEventHandlerProtocol(watchdog_events.FileSystemEventHandler):
        active_venv: qik.venv.Active | None
else:
    import qik.lazy

    watchdog_events = qik.lazy.module("watchdog.events")
    watchdog_observers = qik.lazy.module("watchdog.observers")


@qik.func.cache
def _parse_pydist(path: str) -> None | str:
    match = re.match(r"^(.+)-([^-]+)\.dist-info/RECORD$", path)

    # The pip installation process seems to use "~" as a temporary marker
    # for package names
    if match and not match.group(1).startswith("~"):
        return qik.dep._normalize_pydist_name(match.group(1))


def _make_watchdog_handler(*, runner: Runner) -> QikEventHandlerProtocol:  # pragma: no cover
    """Create the watchdog event handler.

    Watch for both internal project and virtual env changes.

    Note that we create the class in the function to avoid invoking
    a lazy watchdog import.
    """

    class QikEventHandler(watchdog_events.FileSystemEventHandler):
        def __init__(self):
            self.timer: threading.Timer | None = None
            self.changes: set[qik.dep.Dep] = set()
            self.runner = runner
            self.lock = threading.Lock()

        @qik.func.cached_property
        def qik_file_re(self) -> re.Pattern:
            hidden_files = f"({fnmatch.translate('**/.*')})|({fnmatch.translate('.*')})"
            ignored_patterns = "(__pycache__)"
            return re.compile(f"^(?!{hidden_files}|{ignored_patterns}$).*$")

        @qik.func.cached_property
        def cwd(self) -> pathlib.Path:
            return pathlib.Path.cwd()

        @qik.func.cached_property
        def active_venv(self) -> qik.venv.Active | None:
            # Find if an active venv is being used by the runner. IF so, we'll watch it.
            for runnable in self.runner.graph:
                if isinstance(runnable.resolved_venv, qik.venv.Active):
                    return runnable.resolved_venv

        def restart_timer(self, interval: float = 0.1):
            if self.timer is not None:
                self.timer.cancel()

            self.timer = threading.Timer(interval, self.handle_events)
            self.timer.start()

        def on_any_event(self, event):
            if not isinstance(
                event,
                (
                    watchdog_events.FileModifiedEvent,
                    watchdog_events.FileCreatedEvent,
                    watchdog_events.FileDeletedEvent,
                    watchdog_events.FileMovedEvent,
                ),
            ):
                return

            with self.lock:
                src_path = pathlib.Path(event.src_path).resolve()
                try:
                    path = str(src_path.relative_to(self.cwd))
                    if path.endswith("qik.toml"):
                        self.runner.logger.print(
                            f"{path} config changed. Please restart watcher.",
                            emoji="construction",
                            color="red",
                        )
                        sys.exit(1)
                    elif self.qik_file_re.match(path):
                        self.changes.add(qik.dep.Glob(path))
                except ValueError:
                    if self.active_venv:
                        try:
                            path = str(src_path.relative_to(self.active_venv.site_packages_dir))
                            if (pydist := _parse_pydist(path)) and event.event_type == "created":
                                self.changes.add(qik.dep.Pydist(pydist))
                        except ValueError:  # Not part of the venv
                            pass

                self.restart_timer()

        def handle_events(self):
            with self.lock:
                if self.changes:
                    if len(self.changes) <= 5:
                        changes = ", ".join(str(c) for c in self.changes)
                        summary = f"Detected changes in {changes}."
                    else:
                        summary = f"Detected {len(self.changes)} changes."

                    self.runner.logger.print(summary, emoji="eyes", color="cyan", event="changes")
                    self.runner.exec(changes=self.changes)
                    self.changes = set()

    return QikEventHandler()  # type: ignore


def start(runner: Runner):  # pragma: no cover
    observer = watchdog_observers.Observer()
    handler = _make_watchdog_handler(runner=runner)
    observer.schedule(handler, ".", recursive=True)
    if handler.active_venv:
        observer.schedule(handler, str(handler.active_venv.site_packages_dir), recursive=True)

    observer.start()
    try:
        while observer.is_alive():
            observer.join(1)
    finally:
        observer.stop()



"""Manage global run context."""

from __future__ import annotations

import contextlib
import contextvars
import os
import threading
from typing import TYPE_CHECKING, Any, Iterator, Literal, TypeVar, overload

import msgspec

import qik.arch
import qik.conf
import qik.errors
import qik.func as qik_func
import qik.unset

if TYPE_CHECKING:
    from qik.runnable import Runnable
    from qik.runner import Runner


_RUNNER: Runner | None = None
_RUNNABLE: contextvars.ContextVar[Runnable | None] = contextvars.ContextVar(
    "_RUNNABLE", default=None
)
_CURR_WORKER_ID: int = 0
_WORKER_IDS: dict[int, int] = {}
_WORKER_LOCK = threading.Lock()
_WORKER_ID: contextvars.ContextVar[int | None] = contextvars.ContextVar("_WORKER_ID", default=None)


def runner() -> Runner:
    """Get the current runner."""
    if not _RUNNER:
        raise RuntimeError("No runner set in context.")

    return _RUNNER


@contextlib.contextmanager
def set_runner(runner: Runner) -> Iterator[None]:
    """Set the current runner."""
    global _RUNNER
    old_runner = _RUNNER
    _RUNNER = runner

    try:
        yield
    finally:
        _RUNNER = old_runner


def runnable() -> Runnable:
    """Get the current runnable."""
    if runnable := _RUNNABLE.get():
        return runnable
    else:
        raise RuntimeError("No runnable set in context.")


@contextlib.contextmanager
def set_runnable(runnable: Runnable) -> Iterator[None]:
    """Set the current runnable."""
    old_runnable = _RUNNABLE.set(runnable)
    try:
        yield
    finally:
        _RUNNABLE.reset(old_runnable)


def worker_id() -> int:
    """Get the current worker ID."""
    if worker_id := _WORKER_ID.get():
        return worker_id
    else:
        raise RuntimeError("No worker ID set in context.")


@contextlib.contextmanager
def set_worker_id() -> Iterator[None]:
    """Set the current worker ID."""
    global _CURR_WORKER_ID, _WORKER_IDS
    thread_ident = threading.get_ident()

    if not _WORKER_IDS.get(thread_ident):
        with _WORKER_LOCK:
            _CURR_WORKER_ID += 1
            _WORKER_IDS[thread_ident] = _CURR_WORKER_ID

    old_worker_id = _WORKER_ID.set(_WORKER_IDS[thread_ident])
    try:
        yield
    finally:
        _WORKER_ID.reset(old_worker_id)


class QikCtx(msgspec.Struct, forbid_unknown_fields=True, rename="kebab", dict=True):
    # Runtime behavior
    isolated: bool = False
    watch: bool = False
    force: bool = False
    ls: bool = False
    workers: int = msgspec.field(default_factory=lambda: os.cpu_count() or 1)
    fail: bool = False
    verbosity: int = 1

    # Selectors
    since: str | None = None
    commands: list[str] = []
    modules: list[str] = []
    spaces: list[str] = []
    cache_status: qik.conf.CacheStatus | None = None
    caches: list[str] = []
    tags: list[str] = []
    logger: Literal["progress", "stdout"] | None = None

    @qik_func.cached_property
    def arch(self) -> qik.arch.ArchType:
        return qik.arch.get()


@qik_func.cache
def _var_struct(namespace: qik.conf.CtxNamespace | None = None) -> type[msgspec.Struct]:
    """Get a struct of vars for a module."""
    if namespace == "qik":
        return QikCtx
    else:
        proj = qik.conf.project()
        return msgspec.defstruct(
            "VarStruct",
            [(v.name, v.pytype, v.default) for v in proj.ctx_vars.values()],  # type: ignore
            forbid_unknown_fields=True,
            rename="kebab",
        )


@overload
def by_namespace(name: qik.conf.CtxNamespace, /) -> QikCtx: ...


@overload
def by_namespace(name: None, /) -> msgspec.Struct: ...


def by_namespace(name: qik.conf.CtxNamespace | None, /) -> QikCtx | msgspec.Struct:
    return _by_namespace(name)


@qik_func.cache
def _by_namespace(name: qik.conf.CtxNamespace | None) -> QikCtx | msgspec.Struct:
    """Get context for a namesapce."""
    env_prefix = f"{name}__" if name else ""
    namespace_prefix = f"{name}." if name else ""
    var_struct = _var_struct(name)
    parsed = msgspec.convert({}, type=var_struct)

    def _get_val(var_name: str, type_val: str | type) -> qik.conf.VarType | qik.unset.UnsetType:
        str_type = type_val.__name__ if isinstance(type_val, type) else str(type_val)
        env_key = f"{env_prefix}{var_name}".upper()
        env_setting = os.environ.get(env_key)
        if env_setting is None:
            return getattr(parsed, var_name)

        match str_type:
            case "str" | "str | None" | "qik.conf.CacheStatus | None":
                return env_setting
            case "int" | "int | None":
                try:
                    return int(env_setting)
                except ValueError as exc:
                    raise qik.errors.EnvCast(
                        f'Unable to cast env ctx {env_key} value "{env_setting}" as int'
                    ) from exc
            case "bool" | "bool | None":
                if env_setting.lower().strip() in ("yes", "true", "1", "no", "false", "0"):
                    return env_setting.lower().strip() in ["yes", "true", "1"]
                else:
                    raise qik.errors.EnvCast(
                        f'Unable to cast env ctx {env_key} value "{env_setting}" as bool.'
                    )
            case "list[str]":
                return env_setting.split(",")
            case other:
                raise AssertionError(f"Unexpected ctx var type: {other}")

    for var_name, var_type in var_struct.__annotations__.items():
        setattr(parsed, var_name, _get_val(var_name, var_type))
        if isinstance(getattr(parsed, var_name), qik.unset.UnsetType):
            raise qik.errors.CtxValueNotFound(
                f'No value supplied for "{namespace_prefix}{var_name}" ctx.'
            )

    return parsed


@contextlib.contextmanager
def set_vars(
    namespace: qik.conf.CtxNamespace,
    **vars: qik.conf.VarType | qik.unset.UnsetType,
) -> Iterator[None]:
    curr_vars = by_namespace(namespace)
    old_vars = msgspec.structs.replace(curr_vars)
    for name, val in vars.items():
        if not isinstance(val, qik.unset.UnsetType):
            setattr(curr_vars, name, val)

    try:
        yield
    finally:
        for name in old_vars.__struct_fields__:
            setattr(curr_vars, name, getattr(old_vars, name))


class _NamespaceCtx:
    """A utility class for accessing namespaced context as an object."""

    def __init__(self, namespace: qik.conf.CtxNamespace | None):
        self._namespace: qik.conf.CtxNamespace | None = namespace
        self._prefix = f"{namespace}." if namespace else ""

    @qik_func.cached_property
    def _namespace_ctx(self) -> msgspec.Struct | QikCtx:
        return by_namespace(self._namespace)

    def __getattr__(self, key: str) -> qik.conf.VarType:
        try:
            return getattr(self._namespace_ctx, key)
        except AttributeError as exc:
            raise qik.errors.UnconfiguredCtx(f'Ctx "{self._prefix}{key}" not configured') from exc


class Ctx:
    """A singleton class for accessing context as an object."""

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __getattribute__(self, name: str) -> Any:
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return getattr(self.project, name)

    @qik_func.cached_property
    def project(self) -> _NamespaceCtx:
        return _NamespaceCtx(None)

    @qik_func.cached_property
    def qik(self) -> _NamespaceCtx:
        return _NamespaceCtx("qik")


_CTX = Ctx()
FMT_STR_T = TypeVar("FMT_STR_T", bound=str | qik.unset.UnsetType | None)


def format(val: FMT_STR_T, module: qik.conf.ModuleLocator | None = None) -> FMT_STR_T:
    """Formats a string with context and other variables"""
    return val.format(module=module, ctx=_CTX) if isinstance(val, str) else val



from __future__ import annotations

import builtins
import pathlib
import shutil
from typing import IO, TYPE_CHECKING, Literal, TypeAlias

import qik.conf

if TYPE_CHECKING:
    PathLike: TypeAlias = pathlib.Path | str

_builtin_open = builtins.open


def make_parent_dirs(path: pathlib.Path) -> None:
    """Create parent dirs of a path.

    Ensures we properly create the structure of our local
    qik cache.
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    if str(path.resolve()).startswith(str(qik.conf.priv_work_dir())):
        # Ensure our project cache directory always has a .gitignore
        (qik.conf.priv_work_dir() / ".gitignore").write_text("*")


def open(path: PathLike, mode: Literal["w"] = "w") -> IO[str]:
    try:
        return _builtin_open(path, "w")
    except FileNotFoundError:
        make_parent_dirs(path)
        return _builtin_open(path, "w")


def copy(src: PathLike, dest: PathLike) -> None:
    """Copy a file, trying again if the parent dirs don't exist."""
    src_path = pathlib.Path(src)
    dest_path = pathlib.Path(dest)
    try:
        shutil.copy(src_path, dest_path)
    except FileNotFoundError:
        make_parent_dirs(dest_path)
        shutil.copy(src_path, dest_path)


def write(location: PathLike, val: bytes | str) -> None:
    """Write bytes to a file."""
    location_path = pathlib.Path(location)
    writer = location_path.write_text if isinstance(val, str) else location_path.write_bytes
    try:
        writer(val)
    except FileNotFoundError:
        make_parent_dirs(location_path)
        writer(val)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Iterator

import msgspec

import qik.conf
import qik.dep
import qik.errors
import qik.file
import qik.func
import qik.hash
import qik.pygraph.conf
import qik.pygraph.core
import qik.pygraph.utils
import qik.runnable
import qik.space
import qik.unset

if TYPE_CHECKING:
    import grimp.exceptions as grimp_exceptions
else:
    import qik.lazy

    grimp_exceptions = qik.lazy.module("grimp.exceptions")


class GraphConfDep(qik.dep.Val, frozen=True, dict=True):
    val: str = "pygraph"  # The value is dynamic based on plugin name
    file: str = ""  # The file is dynamic based on config location

    @qik.func.cached_property
    def watch(self) -> list[str]:
        return [qik.conf.location().name]

    @qik.func.cached_property
    def vals(self) -> list[str]:  # type: ignore
        conf = qik.pygraph.conf.get()
        return [str(msgspec.json.encode(conf))]


@qik.func.cache
def _graph_conf_dep() -> qik.dep.Val:
    """Serialize the graph config so it can be included as a dependency."""
    return GraphConfDep()


def build_cmd(runnable: qik.runnable.Runnable) -> tuple[int, str]:
    """Build the import graph."""
    try:
        graph = qik.pygraph.core.build()
    except grimp_exceptions.SourceSyntaxError as exc:
        return 1, f"Cannot build import graph - {exc}"

    path = qik.pygraph.utils.graph_path(rel=False)

    serialized = msgspec.json.encode(graph)
    qik.file.write(path, serialized)
    return 0, ""


def lock_cmd(runnable: qik.runnable.Runnable) -> tuple[int, str]:
    pyimport = runnable.args.get("pyimport")
    if not pyimport:
        raise AssertionError("Unexpected qik.pygraph.lock runnable.")

    graph = load_graph()
    # TODO: Better error if the module doesn't exist
    upstream = graph.upstream_modules(pyimport, idx=False)
    root = graph.modules[graph.modules_idx[pyimport]]
    distributions = runnable.resolved_venv.packages_distributions()

    def _gen_upstream_globs() -> Iterator[str]:
        for module in [root, *upstream]:
            if module.is_internal:
                path = module.imp.replace(".", "/")
                yield f"{path}/**.py"
                yield f"{path}.py"

    def _gen_upstream_pydists() -> Iterator[str]:
        for module in upstream:
            if not module.is_internal:
                try:
                    yield from (pydist for pydist in distributions[module.imp] if pydist)
                except KeyError as exc:
                    raise qik.errors.ModuleDistributionNotFound(
                        f'No distribution found for module "{module.imp}"'
                    ) from exc

    runnable.store_deps(
        qik.pygraph.utils.lock_path(pyimport, runnable.space, rel=False),
        globs=sorted(_gen_upstream_globs()),
        pydists=sorted(_gen_upstream_pydists()),
    )
    return 0, ""


def _generate_fence_regex(valid_imports) -> re.Pattern:
    valid_paths = "|".join(re.escape(imp) + r"(?:\.[^:]+)?" for imp in valid_imports)
    pattern = rf"^({valid_paths}):(?!({valid_paths})(?:\.|$)).*$"
    return re.compile(pattern, re.MULTILINE)


def check_cmd(runnable: qik.runnable.Runnable) -> tuple[int, str]:
    graph = load_graph()
    # TODO: Handle custom PYTHONPATH env var
    fence_pyimports = runnable.resolved_space.fence_pyimports
    imps: set[tuple[qik.pygraph.core.Module, qik.pygraph.core.Module]] = set().union(
        *(graph.upstream_imports(imp) for imp in fence_pyimports)
    )

    def _fmt_violation(imp: str) -> str:
        """Create a human readable violation message."""
        src, dest = imp.split(":")
        return f"{src} imports {dest}"

    internal_imps = "\n".join(f"{src.imp}:{dest.imp}" for src, dest in imps if dest.is_internal)
    internal_fence_re = _generate_fence_regex(fence_pyimports)
    internal_violations = [
        _fmt_violation(violation.group())
        for violation in re.finditer(internal_fence_re, internal_imps)
    ]

    external_imps = "\n".join(
        f"{src.imp}:{dest.imp}" for src, dest in imps if not dest.is_internal
    )
    external_fence_re = _generate_fence_regex(
        (*fence_pyimports, *runnable.resolved_venv.packages_distributions())
    )
    external_violations = [
        _fmt_violation(violation.group())
        for violation in re.finditer(external_fence_re, external_imps)
    ]

    ret_code = 0 if not internal_violations and not external_violations else 1
    if ret_code == 1:
        stdout = ""
        if internal_violations:
            stdout += (
                f'Found {len(internal_violations)} internal import violations for "{runnable.space}" space:\n'
                + "\n".join(internal_violations)
            ) + "\n\n"
        if external_violations:
            stdout += (
                f'Found {len(external_violations)} external import violations for "{runnable.space}" space:\n'
                + "\n".join(external_violations)
            )
    else:
        stdout = (
            f'No import violations found across {len(imps)} imports in "{runnable.space}" space'
        )

    return ret_code, stdout.strip()


def build_cmd_factory(
    cmd: str, conf: qik.conf.Cmd, **args: str
) -> dict[str, qik.runnable.Runnable]:
    build_graph_cmd_name = qik.pygraph.utils.build_cmd_name()
    pygraph_conf = qik.pygraph.conf.get()
    runnable = qik.runnable.Runnable(
        name=qik.runnable.fmt_name(build_graph_cmd_name),
        cmd=build_graph_cmd_name,
        val="qik.pygraph.cmd.build_cmd",
        shell=False,
        deps=[*qik.dep.base(), qik.dep.Glob("**.py"), _graph_conf_dep()],
        artifacts=[str(qik.pygraph.utils.graph_path())],
        cache=pygraph_conf.resolved_build_cache,
        cache_when="success",
    )
    return {runnable.name: runnable}


def lock_cmd_factory(
    cmd: str, conf: qik.conf.Cmd, **args: str
) -> dict[str, qik.runnable.Runnable]:
    if "pyimport" not in args or "space" not in args:
        raise qik.errors.ArgNotSupplied(
            '"pyimport" and "space" args are required for qik.pygraph.lock command.'
        )

    pyimport = args["pyimport"]
    space = args["space"]
    cmd_name = qik.pygraph.utils.lock_cmd_name()
    name = qik.runnable.fmt_name(cmd_name, space=space, pyimport=pyimport)
    pygraph_conf = qik.pygraph.conf.get()
    artifact = str(qik.pygraph.utils.lock_path(pyimport, space))
    runnable = qik.runnable.Runnable(
        name=name,
        cmd=cmd_name,
        val="qik.pygraph.cmd.lock_cmd",
        shell=False,
        deps=[
            *qik.dep.base(),
            qik.dep.Cmd(qik.pygraph.utils.build_cmd_name()),
            qik.dep.Load(artifact, default=["**.py"]),
            _graph_conf_dep(),
        ],
        artifacts=[artifact],
        cache=pygraph_conf.resolved_lock_cache,
        cache_when="success",
        space=space,
        args={"pyimport": pyimport},
    )
    return {runnable.name: runnable}


def check_cmd_factory(
    cmd: str, conf: qik.conf.Cmd, **args: str
) -> dict[str, qik.runnable.Runnable]:
    cmd_name = qik.pygraph.utils.check_cmd_name()
    pygraph_conf = qik.pygraph.conf.get()

    runnables = (
        qik.runnable.Runnable(
            name=qik.runnable.fmt_name(cmd_name, space=space),
            cmd=cmd_name,
            val="qik.pygraph.cmd.check_cmd",
            shell=False,
            deps=[
                qik.dep.Cmd(qik.pygraph.utils.build_cmd_name()),
                *(qik.dep.Glob(fence_path) for fence_path in qik.space.load(space).fence_paths),
            ],
            cache=pygraph_conf.resolved_check_cache,
            cache_when="success",
            space=space,
        )
        for space, space_conf in qik.conf.project().resolved_spaces.items()
        if space_conf.fence
    )
    return {runnable.name: runnable for runnable in runnables}


@qik.func.per_run_cache
def load_graph() -> qik.pygraph.core.Graph:
    """Load the graph."""
    return msgspec.json.decode(
        qik.pygraph.utils.graph_path(rel=False).read_bytes(), type=qik.pygraph.core.Graph
    )



"""Core hashing functions."""

from __future__ import annotations

import collections
import re
import subprocess
from typing import TYPE_CHECKING

import xxhash

import qik.conf
import qik.errors
import qik.func
import qik.shell

if TYPE_CHECKING:
    import qik.dep as run_deps
    import qik.venv


def _run_with_stdin(cmd: list[str], stdin_data: str) -> list[str]:
    """Run a command with data piped via stdin, avoiding ARG_MAX limits."""
    result = subprocess.run(
        cmd,
        input=stdin_data,
        text=True,
        capture_output=True,
        check=True,
    )
    return [line for line in result.stdout.strip().split("\n") if line]


def globs(*vals: run_deps.Glob | str) -> str:
    """Compute a hash string of glob patterns.

    Use git ls-files, tracking both added and modified file hashes.

    If local ._qik/artifacts/ exists, we'll manually hash these.
    """
    if not vals:
        return ""

    globs = {f"{glob}" for glob in vals}

    priv_artifact_patterns = sorted(
        f"'{glob[16:]}'" for glob in globs if glob.startswith("._qik/artifacts/")
    )
    priv_artifact_hash = ""
    if priv_artifact_patterns:
        priv_artifact_hash = qik.shell.exec(
            f"git -C ._qik/artifacts hash-object {' '.join(priv_artifact_patterns)}",
            check=True,
        ).stdout

    repo_patterns = sorted(
        glob for glob in globs if not glob.startswith("._qik/artifacts/")
    )
    repo_hash = ""
    if repo_patterns:
        # Pipe patterns via stdin to xargs to avoid ARG_MAX limits
        patterns_input = "\n".join(repo_patterns)
        git_ls_lines = _run_with_stdin(
            ["xargs", "git", "ls-files", "-cms"],
            patterns_input,
        )
        git_ls_lines_split = [
            (path, sha)
            for _, sha, _, path in (re.split(r"\s+", line, maxsplit=3) for line in git_ls_lines)
        ]
        hashes = dict(git_ls_lines_split)

        # Files that are modified appear twice. Manually compute their hashes
        path_counts = collections.Counter(line[0] for line in git_ls_lines_split)
        modified = [path for path, count in path_counts.items() if count > 1]
        if modified:
            modified_paths_str = "\n".join(modified)
            try:
                modified_hashes_lines = _run_with_stdin(
                    ["xargs", "git", "hash-object"],
                    modified_paths_str,
                )
            except subprocess.CalledProcessError:
                # If there are issues with the first command, it likely means a file does
                # not exist. Do the suboptimal strategy here, hashing individual files
                # while printing zeroes for files that no longer exist.
                modified_hashes_lines = []
                for path in modified:
                    try:
                        result = subprocess.run(
                            ["git", "hash-object", path],
                            text=True,
                            capture_output=True,
                            check=True,
                        )
                        modified_hashes_lines.append(result.stdout.strip())
                    except subprocess.CalledProcessError:
                        modified_hashes_lines.append("0000000000000000000000000000000000000000")

            for i, name in enumerate(modified):
                # TODO: Occasionally we can get a list index out of range error here. It seems
                # to be due to a race condition. Re-raise it for now with more context in hopes
                # of better debugging it.
                try:
                    hashes[name] = modified_hashes_lines[i]
                except IndexError as e:
                    raise RuntimeError(f"Unexpected error when computing hash. {modified}") from e

        repo_hash = "".join(f"{name}{hash}" for name, hash in sorted(hashes.items()))

    return xxhash.xxh128_hexdigest(priv_artifact_hash + repo_hash)


def pydists(*vals: str, venv: qik.venv.Venv) -> str:
    return xxhash.xxh128_hexdigest(
        "".join(f"{pydist}{venv.version(pydist)}" for pydist in sorted(vals))
    )


def strs(*vals: str) -> str:
    return xxhash.xxh128_hexdigest("".join(sorted(vals)))


def val(input: str | bytes) -> str:
    return xxhash.xxh128_hexdigest(input)



import importlib
from typing import Any, Callable, Final, Generic, TypeVar

import qik.func

T = TypeVar("T")
_MODULE_TO_OPT_INSTALL: Final = {
    "watchdog": "watch",
    "grimp": "pygraph",
    "rustworkx": "pygraph",
    "boto3": "s3",
}


class object(Generic[T]):
    def __init__(self, loader: Callable[[], T]) -> None:
        self._loader = loader

    @qik.func.cached_property
    def _object(self) -> T:
        return self._loader()

    def __enter__(self, *args, **kwargs) -> Any:
        return self._object.__enter__(*args, **kwargs)  # type: ignore

    def __exit__(self, *args, **kwargs) -> Any:
        return self._object.__exit__(*args, **kwargs)  # type: ignore

    def __getattr__(self, name: str) -> Any:
        return getattr(self._object, name)


class module:
    def __init__(self, modname: str) -> None:
        self._modname = modname
        self._mod = None

    def __getattr__(self, attr: str) -> Any:
        try:
            return getattr(self._mod, attr)
        except Exception:
            if self._mod is None:
                try:
                    self._mod = importlib.import_module(self._modname)
                except ModuleNotFoundError as exc:
                    err_msg = f"{self._modname} could not be imported."
                    package = self._modname.split(".", 1)[0]

                    if opt_install := _MODULE_TO_OPT_INSTALL.get(package):
                        err_msg += f' Do "pip install qik[{opt_install}]" or install the {package} package to use qik.'
                    else:
                        err_msg += f" Install the {package} package to use qik."
                    raise ModuleNotFoundError(err_msg) from exc
            else:
                raise

        return getattr(self._mod, attr)



from __future__ import annotations

import collections
import contextlib
import os
import pathlib
from typing import IO, TYPE_CHECKING, Iterator, Literal, TypeAlias

import msgspec

import qik.conf
import qik.console
import qik.ctx
import qik.file
import qik.func

if TYPE_CHECKING:
    import rich.live as rich_live
    import rich.progress as rich_progress
    import rich.table as rich_table
    import rich.text as rich_text
    from rich.console import RenderableType

    from qik.cache import Entry
    from qik.runnable import Result, Runnable
    from qik.runner import Graph

    Event: TypeAlias = Literal["start", "changes", "finish", "exception", "output"]
    from qik.console import Color, Emoji
else:
    import qik.lazy

    rich_progress = qik.lazy.module("rich.progress")
    rich_style = qik.lazy.module("rich.style")
    rich_table = qik.lazy.module("rich.table")
    rich_text = qik.lazy.module("rich.text")
    rich_live = qik.lazy.module("rich.live")

RunStatus: TypeAlias = Literal["success", "failed", "skipped", "pending"]


@qik.func.cache
def out_dir() -> pathlib.Path:
    # Note, relative_to(walk_up=True) is supported in python 3.12.
    return pathlib.Path(os.path.relpath(qik.conf.priv_work_dir() / "out", os.getcwd()))


class Manifest(msgspec.Struct):
    cmd_status: dict[str, RunStatus]
    runnables: list[str]
    runnable_status: dict[str, RunStatus]
    cached_runnables: list[str]


class Stats:
    def __init__(self, graph: Graph):
        self.manifest_path = out_dir() / "manifest.json"

        # Command-level status information
        self.cmd_finished: dict[str, int] = collections.defaultdict(int)
        self.cmd_running: dict[str, bool] = collections.defaultdict(bool)
        self.cmd_total: dict[str, int] = collections.defaultdict(int)
        for runnable in graph:
            self.cmd_total[runnable.cmd] += 1

        self.cmd_status: dict[str, RunStatus] = collections.defaultdict(lambda: "pending")

        # Runnable-level status information
        self.runnables = [runnable.name for runnable in graph]
        self.runnable_status: dict[str, RunStatus] = {
            runnable.name: "pending" for runnable in graph
        }
        self.cached_runnables = []

    @property
    def run_failed(self) -> bool:
        return any(status == "failed" for status in self.cmd_status.values())

    def start(self, runnable: Runnable) -> None:
        self.cmd_running[runnable.cmd] = True

    def finish(
        self, runnable: Runnable, *, result: Result | None, cache_entry: Entry | None
    ) -> None:
        # Track runnable status
        if not result:
            self.runnable_status[runnable.name] = "skipped"
        elif result.code != 0:
            self.runnable_status[runnable.name] = "failed"
        else:
            self.runnable_status[runnable.name] = "success"

        if cache_entry:
            self.cached_runnables.append(runnable.name)

        # Track command status
        self.cmd_finished[runnable.cmd] += 1

        if self.cmd_finished[runnable.cmd] >= self.cmd_total[runnable.cmd]:
            self.cmd_running[runnable.cmd] = False

        if self.cmd_status[runnable.cmd] == "pending":
            if not result:
                self.cmd_status[runnable.cmd] = "skipped"
            elif result.code != 0:
                self.cmd_status[runnable.cmd] = "failed"
            elif self.cmd_finished[runnable.cmd] >= self.cmd_total[runnable.cmd]:
                self.cmd_status[runnable.cmd] = "success"

    def write(self) -> None:
        qik.file.write(
            self.manifest_path,
            msgspec.json.encode(
                Manifest(
                    runnables=self.runnables,
                    cmd_status=self.cmd_status,
                    cached_runnables=self.cached_runnables,
                    runnable_status=self.runnable_status,
                )
            ),
        )


class Logger:
    def __init__(self):
        self._reset_state()
        self.num_runs = 0

    def _reset_state(self) -> None:
        self.graph = None
        log_files: dict[str, IO[str]] = getattr(self, "_log_files", {})
        for file in log_files.values():
            file.close()
        self._log_files = {}
        self._stats: Stats | None = None

    def _get_log_file(self, runnable: Runnable) -> IO[str]:
        if runnable.name not in self._log_files:
            path = out_dir() / f"{runnable.name}.out"
            self._log_files[runnable.name] = qik.file.open(path, "w")

        return self._log_files[runnable.name]

    @property
    def stats(self) -> Stats:
        if not self._stats:
            raise AssertionError("Stats not initialized")

        return self._stats

    @stats.setter
    def stats(self, stats: Stats) -> None:
        self._stats = stats

    @contextlib.contextmanager
    def run(self, graph: Graph) -> Iterator[None]:
        self.graph = graph
        self.stats = Stats(graph)
        self.stats.write()
        self.handle_run_started()
        try:
            yield
        finally:
            self.handle_run_finished()
            self.stats.write()
            self._reset_state()

    def print(
        self,
        msg: str,
        runnable: Runnable | None = None,
        cache_entry: Entry | None = None,
        result: Result | None = None,
        event: Event | None = None,
        emoji: Emoji | None = None,
        color: Color | None = None,
    ) -> None:
        if runnable:
            if event == "output":
                if msg := msg.strip():
                    self._get_log_file(runnable).write(msg)
            elif event == "start":
                self.stats.start(runnable)
            elif event == "finish":
                self.stats.finish(runnable, result=result, cache_entry=cache_entry)

        self.handle_output(
            msg,
            runnable=runnable,
            cache_entry=cache_entry,
            result=result,
            event=event,
            emoji=emoji,
            color=color,
        )

    def handle_run_started(self) -> None:
        if self.num_runs and self.graph:
            qik.console.rule(color="white")

        self.num_runs += 1

    def handle_run_finished(self) -> None:
        pass

    def handle_output(
        self,
        msg: str,
        runnable: Runnable | None = None,
        cache_entry: Entry | None = None,
        result: Result | None = None,
        event: Event | None = None,
        emoji: Emoji | None = None,
        color: Color | None = None,
    ) -> None:
        pass


class Stdout(Logger):
    """Logs command results to stdout."""

    def handle_output(
        self,
        msg: str,
        runnable: Runnable | None = None,
        cache_entry: Entry | None = None,
        result: Result | None = None,
        event: Event | None = None,
        emoji: Emoji | None = None,
        color: Color | None = None,
    ) -> None:
        """Print output from a runnable."""
        if event == "output":
            if msg := msg.strip():
                qik.console.print(msg, emoji=emoji, color=color, style=None)
        elif event == "start" or (event == "finish" and not result):
            qik.console.rule(msg, emoji=emoji, color=color)
        else:
            kwargs = {"overflow": "ellipsis", "no_wrap": True} if event == "finish" else {}
            qik.console.print(msg, emoji=emoji, color=color, **kwargs)

        if event == "exception":
            qik.console.print_exception()


class HideableBarColumn(rich_progress.BarColumn):
    def render(self, task) -> RenderableType:  # type: ignore
        if task.fields.get("show_progress", True):
            return super().render(task)
        return rich_text.Text("")


class HideableSpinnerColumn(rich_progress.SpinnerColumn):
    def render(self, task) -> RenderableType:
        if task.fields.get("show_progress", True):
            return super().render(task)
        return rich_text.Text("")


class Progress(Logger):
    """Logs live progress bars and prints output at end."""

    def __init__(self):
        super().__init__()
        self.status: RunStatus = "pending"

    def get_status_text(self) -> str:
        status_text = f"Output in [bold]{out_dir()}"
        if self.status == "pending":
            status_text = f":heavy_minus_sign-emoji: [cyan]{status_text}"
        elif self.status == "success":
            status_text = f":white_check_mark-emoji: [green]{status_text}"
        elif self.status == "failed":
            status_text = f":broken_heart-emoji: [red]{status_text}"

        return status_text

    def generate_table(self):
        table = rich_table.Table.grid(padding=(0, 1))
        table.add_row(self.progress)

        status_text = f"Output in [bold]{out_dir()}"
        if self.status == "pending":
            status_text = f":heavy_minus_sign-emoji: [cyan]{status_text}"
        elif self.status == "success":
            status_text = f":white_check_mark-emoji: [green]{status_text}"
        elif self.status == "failed":
            status_text = f":broken_heart-emoji: [red]{status_text}"

        if self.status == "pending":
            table.add_row(self.get_status_text())

        return table

    def handle_run_started(self) -> None:
        super().handle_run_started()

        self.status = "pending"
        self.progress = rich_progress.Progress(
            rich_progress.TextColumn("{task.description}"),
            HideableBarColumn(complete_style="cyan"),
            HideableSpinnerColumn(style="cyan"),
            expand=True,
        )

        self.cmds = {
            cmd: self.progress.add_task(
                f":heavy_minus_sign-emoji: [dim]{cmd}",
                total=self.stats.cmd_total[cmd],
                show_progress=False,
            )
            for cmd in sorted(self.stats.cmd_total)
        }
        self.captured: dict[str, list[str]] = collections.defaultdict(list)

        self.live = rich_live.Live(
            self.generate_table(), refresh_per_second=10, console=qik.console.get()
        )
        self.live.start()

    def handle_output(
        self,
        msg: str,
        runnable: Runnable | None = None,
        cache_entry: Entry | None = None,
        result: Result | None = None,
        event: Event | None = None,
        emoji: Emoji | None = None,
        color: Color | None = None,
    ) -> None:
        """Print output from a runnable."""
        if event == "start" and runnable:
            self.progress.update(
                self.cmds[runnable.cmd],
                description=f":construction-emoji: [bold blue]{runnable.cmd}",
                show_progress=True,
            )
        elif event == "finish" and runnable:
            match self.stats.cmd_status[runnable.cmd]:
                case "skipped":
                    status = ":heavy_minus_sign-emoji:[yellow]"
                case "failed":
                    status = ":broken_heart-emoji:[red]"
                case "success":
                    status = ":white_check_mark-emoji:[green]"
                case "pending":
                    status = ":construction-emoji:[bold blue]"

            self.progress.update(
                self.cmds[runnable.cmd],
                advance=1,
                description=f"{status} {runnable.cmd}",
                show_progress=self.stats.cmd_running[runnable.cmd],
            )
            self.live.update(self.generate_table())
        elif event == "exception":
            qik.console.print(msg, emoji=emoji, color=color)
            qik.console.print_exception()
        elif event != "output":
            qik.console.print(msg, emoji=emoji, color=color)
        elif runnable:
            self.captured[runnable.name].append(msg)

    def handle_run_finished(self) -> None:
        self.status = "failed" if self.stats.run_failed else "success"
        self.live.update(self.generate_table())
        self.live.stop()

        verbosity = qik.ctx.by_namespace("qik").verbosity
        if verbosity:
            cached_runnables = self.stats.cached_runnables

            for name in sorted(self.stats.runnables):
                match self.stats.runnable_status[name]:
                    case "success":
                        emoji = (
                            "fast-forward_button"
                            if name in cached_runnables
                            else "white_check_mark"
                        )
                        color = "green"
                        show = True if verbosity > 1 else False
                        output = "\n".join(line.strip() for line in self.captured[name])
                    case "failed":
                        emoji = (
                            "fast-forward_button" if name in cached_runnables else "broken_heart"
                        )
                        color = "red"
                        show = True if verbosity > 0 else False
                        output = "\n".join(line.strip() for line in self.captured[name])
                    case _:
                        emoji = "heavy_minus_sign"
                        color = "yellow"
                        show = True if verbosity > 1 else False
                        output = "[dim][italic]Skipped"

                if show:
                    qik.console.rule(f":{emoji}-emoji: [{color}]{name}", color=color)
                    output = output or "[dim][italic]No output"
                    qik.console.print(output)

        qik.console.print(self.get_status_text())



import subprocess
from typing import Literal, overload

import qik.conf
import qik.ctx


@overload
def exec(cmd: str, /, *, check: bool = ...) -> subprocess.CompletedProcess[str]: ...


@overload
def exec(cmd: str, /, *, lines: Literal[True], check: bool = ...) -> list[str]: ...


@overload
def exec(
    cmd: str, /, *, lines: Literal[False], check: bool = ...
) -> subprocess.CompletedProcess[str]: ...


def exec(
    cmd: str, /, *, lines: bool = False, check: bool = False, env: dict[str, str] | None = None
) -> subprocess.CompletedProcess[str] | list[str]:
    """Run a shell commmand.

    If lines=True, return stdout as parsed lines.
    """
    result = subprocess.run(
        cmd,
        shell=True,
        text=True,
        capture_output=True,
        check=check,
        env=env,
        cwd=qik.conf.root(),
    )
    if not lines:
        return result
    else:
        stdout = result.stdout.strip()
        return stdout.split("\n") if stdout else []


def popen(cmd: str, env: dict[str, str] | None = None) -> subprocess.Popen:
    return subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        shell=True,
        bufsize=1,
        universal_newlines=True,
        env=env,
        cwd=qik.conf.root(),
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import qik.conf
import qik.func
import qik.pygraph.qikplugin

if TYPE_CHECKING:
    from qik.pygraph.qikplugin import PygraphPluginConf


@qik.func.cache
def get() -> PygraphPluginConf:
    """Get the pygraph config."""
    return qik.conf.plugin("qik.pygraph")  # type: ignore



from __future__ import annotations

import platform
from typing import TYPE_CHECKING, Literal, TypeAlias

if TYPE_CHECKING:
    ArchType: TypeAlias = Literal[
        "win-64",
        "win-32",
        "linux-64",
        "linux-aarch64",
        "linux-ppc64le",
        "linux-32",
        "osx-arm64",
        "osx-64",
    ]


def get() -> ArchType:
    """Get the architecture of a machine."""
    system = platform.system()
    arch = platform.architecture()[0]
    machine = platform.machine()

    if system == "Windows":
        return "win-64" if arch == "64bit" else "win-32"
    elif system == "Linux":
        if machine == "x86_64":
            return "linux-64"
        elif machine == "aarch64":
            return "linux-aarch64"
        elif machine == "ppc64le":
            return "linux-ppc64le"
        elif arch == "32bit":
            return "linux-32"
    elif system == "Darwin":
        if arch == "64bit":
            return "osx-arm64" if machine == "arm64" else "osx-64"

    return "unknown"



"""Qik project and configuration loading."""

from __future__ import annotations

import collections
import importlib.util
import os.path
import pathlib
import sys
from types import UnionType
from typing import Any, ClassVar, Generator, Literal, TypeAlias, TypeVar, Union

import msgspec.structs
import msgspec.toml

import qik.conf
import qik.errors
import qik.func
import qik.unset

CtxNamespace: TypeAlias = Literal["qik", "project", "modules", "plugins"]
SerializableVarType: TypeAlias = str | bool | int
VarType: TypeAlias = SerializableVarType | list[str]
VAR_T = TypeVar("VAR_T", str, bool, int, list[str], None)
CacheWhen: TypeAlias = Literal["success", "failed", "finished"]
CacheStatus: TypeAlias = Literal["warm", "code"]


# Dynamic objects registered by plugins
_PLUGIN_TYPES: dict[str, dict[str, tuple[type[BasePluggable], str]]] = collections.defaultdict(
    dict
)
# Dynamic conf classes registered by plugins
_CONF_TYPES: dict[str, type[msgspec.Struct]] = {}
# Default venv type
_VENV_TYPE: type[Venv] | None = None


def register_type(plugin_type: type[BasePluggable], factory: str) -> None:
    if plugin_type.plugin_type_name == "venv":
        global _VENV_TYPE
        _VENV_TYPE = plugin_type  # type: ignore

    _PLUGIN_TYPES[plugin_type.plugin_type_name][str(plugin_type.__struct_config__.tag)] = (
        plugin_type,
        factory,
    )


def register_conf(conf_type: type[qik.conf.PluginConf]) -> None:
    _CONF_TYPES[str(conf_type.__struct_config__.tag)] = conf_type


def get_type_factory(conf: BasePluggable) -> str:
    if entry := _PLUGIN_TYPES[conf.plugin_type_name].get(str(conf.__struct_config__.tag)):
        return entry[1]
    else:
        raise qik.errors.InvalidCacheType(
            f'{conf.plugin_type_name.title()} type "{conf.__struct_config__.tag}" not provided by any plugin.'
        )


class Base(
    msgspec.Struct,
    frozen=True,
    omit_defaults=True,
    forbid_unknown_fields=True,
    rename="kebab",
    dict=True,
):
    pass


class BasePluggable(Base, frozen=True):
    plugin_type_name: ClassVar[str]


class Dep(BasePluggable, frozen=True):
    plugin_type_name: ClassVar[str] = "dep"


class GlobDep(Dep, tag="glob", frozen=True):
    pattern: str


class ConstDep(Dep, tag="const", frozen=True):
    val: str


class ValDep(Dep, tag="val", frozen=True):
    key: str
    file: str


class CmdDep(Dep, tag="command", frozen=True):
    name: str
    strict: bool = False
    isolated: bool | qik.unset.UnsetType = qik.unset.UNSET


class PydistDep(Dep, tag="pydist", frozen=True):
    name: str


class LoadDep(Dep, tag="load", frozen=True):
    path: str
    default: list[str] = []


class Cmd(Base, frozen=True):
    exec: str = ""
    deps: list[str | Dep] = []
    artifacts: list[str] = []
    cache: str | qik.unset.UnsetType = qik.unset.UNSET
    cache_when: CacheWhen | qik.unset.UnsetType = qik.unset.UNSET
    factory: str = ""
    hidden: bool = False
    space: str | qik.unset.UnsetType = qik.unset.UNSET
    tags: list[str] = []


class Var(Base, frozen=True):
    name: str
    type: Literal["str", "int", "bool"] = "str"
    default: SerializableVarType | qik.unset.UnsetType = qik.unset.UNSET
    required: bool = True

    def __post_init__(self):
        if not self.required and isinstance(self.default, qik.unset.UnsetType):
            msgspec.structs.force_setattr(self, "default", None)

    @property
    def pytype(self) -> type | UnionType:
        if self.required:
            return __builtins__[self.type]
        else:
            return __builtins__[self.type] | None


class ModuleOrPlugin(Base, frozen=True):
    commands: dict[str, Cmd | str] = {}


class BaseLocator(Base, frozen=True):
    @qik.func.cached_property
    def dir(self) -> pathlib.Path:
        raise NotImplementedError

    @qik.func.cached_property
    def pyimport(self) -> str:
        raise NotImplementedError

    @qik.func.cached_property
    def conf(self) -> ModuleOrPlugin:
        try:
            return msgspec.toml.decode(
                (self.dir / "qik.toml").read_bytes(),
                type=ModuleOrPlugin,
            )
        except FileNotFoundError:
            return ModuleOrPlugin()


class ModuleLocator(BaseLocator, frozen=True):
    name: str
    path: str

    @qik.func.cached_property
    def pyimport(self) -> str:
        return pyimport(self.path)

    @qik.func.cached_property
    def dir(self) -> pathlib.Path:
        # TODO: While this handles most windows paths, it does not handle literal '/'
        # in paths that are escaped (e.g. my\/file/path)
        return pathlib.Path(self.path.replace("/", os.path.sep))


class BasePluginLocator(BaseLocator, frozen=True):
    @qik.func.cached_property
    def dir(self) -> pathlib.Path:
        spec = importlib.util.find_spec(self.pyimport)
        if not spec or not spec.origin:
            raise qik.errors.PluginImport(f'Could not import plugin "{self.pyimport}"')

        return pathlib.Path(spec.origin).parent


# Note - we use forbid_unknown_fields=False because this is used in our first pass
# to load just the plugins from the conf file.
class PluginLocator(BasePluginLocator, frozen=True, forbid_unknown_fields=False):
    pyimport: str  # type: ignore


class PluginConf(BasePluginLocator, frozen=True, tag_field="pyimport"):
    @property
    def pyimport(self) -> str:  # type: ignore
        return str(self.__struct_config__.tag)


class Venv(BasePluggable, frozen=True, tag_field="type", kw_only=True):
    plugin_type_name: ClassVar[str] = "venv"
    install_cmd: ClassVar[str | None] = None
    reqs: str | list[str]
    lock: str | None = None


class SpaceVenv(Venv, frozen=True, tag="space", kw_only=True):
    name: str
    reqs: str | list[str] = []


class ActiveVenv(Venv, frozen=True, tag="active"):
    reqs: str | list[str] = []


class Cache(BasePluggable, frozen=True, tag_field="type"):
    plugin_type_name: ClassVar[str] = "cache"


class SpaceLocator(Base, tag_field="type", frozen=True, tag="space"):
    name: str


class Space(Base, frozen=True):
    root: str | None = None
    modules: list[str | ModuleLocator] = []
    fence: list[str | SpaceLocator] | bool = []
    venv: Venv | str | list[str] | None = None
    dotenv: str | list[str] | None = None

    @qik.func.cached_property
    def modules_by_name(self) -> dict[str, ModuleLocator]:
        module_locators = (
            ModuleLocator(name=m, path=m) if isinstance(m, str) else m for m in self.modules
        )
        return {m.name: m for m in module_locators}

    @qik.func.cached_property
    def modules_by_path(self) -> dict[str, ModuleLocator]:
        return {m.path: m for m in self.modules_by_name.values()}


class Pydist(Base, frozen=True):
    versions: dict[str, str] = {}
    ignore_missing: bool = False
    ignore_missing_modules: bool = False
    modules: dict[str, str] = {}


class PluginsMixin:
    """Parses the root qik.toml file for plugins.

    Allows us to dynamically determine the structure of the config based on installed
    plugins.
    """

    plugins: dict[str, str | PluginLocator] = {}

    @qik.func.cached_property
    def plugins_by_name(self) -> dict[str, PluginLocator]:
        return {
            name: PluginLocator(pyimport=p) if isinstance(p, str) else p
            for name, p in self.plugins.items()
        }

    @qik.func.cached_property
    def plugins_by_pyimport(self) -> dict[str, tuple[str, PluginLocator]]:
        return {p.pyimport: (name, p) for name, p in self.plugins_by_name.items()}


class Plugins(msgspec.Struct, PluginsMixin, frozen=True, forbid_unknown_fields=False):
    """Parses the root qik.toml file for plugins.

    Allows us to dynamically determine the structure of the config based on installed
    plugins.
    """

    plugins: dict[str, str | PluginLocator] = {}


class BaseConf(Base, frozen=True):
    deps: list[str | Dep] = []


class Defaults(Base, frozen=True):
    cache: str | qik.unset.UnsetType = qik.unset.UNSET
    cache_when: CacheWhen | qik.unset.UnsetType = qik.unset.UNSET
    python_path: str = "."
    # Note: "None" is the sentinel value for these. We can't mix UNSET with a list.
    venv: Venv | str | list[str] | None = None
    dotenv: str | list[str] | None = None


class Project(ModuleOrPlugin, PluginsMixin, frozen=True):
    plugins: dict[str, str | PluginLocator] = {}
    ctx: list[str | Var] = []
    caches: dict[str, Cache] = {}
    spaces: dict[str, Space | str | list[str]] = {}
    base: BaseConf = msgspec.field(default_factory=BaseConf)
    defaults: Defaults = msgspec.field(default_factory=Defaults)
    pydist: Pydist = msgspec.field(default_factory=Pydist)

    @qik.func.cached_property
    def resolved_spaces(self) -> dict[str, Space]:
        return {
            name: Space(venv=space) if isinstance(space, str | list) else space
            for name, space in self.spaces.items()
        }

    @qik.func.cached_property
    def ctx_vars(self) -> dict[str, Var]:
        return dict((v, Var(v)) if isinstance(v, str) else (v.name, v) for v in self.ctx)

    @qik.func.cached_property
    def modules_by_name(self) -> dict[str, ModuleLocator]:
        return {
            name: locator
            for space in self.resolved_spaces.values()
            for name, locator in space.modules_by_name.items()
        }

    @qik.func.cached_property
    def modules_by_path(self) -> dict[str, ModuleLocator]:
        return {m.path: m for m in self.modules_by_name.values()}


def _load_plugins(conf: Plugins) -> None:
    """Load plugins and return the Project configuration."""
    for plugin in conf.plugins.values():
        pyimport = plugin if isinstance(plugin, str) else plugin.pyimport

        try:
            importlib.import_module(f"{pyimport}.qikplugin")
        except ModuleNotFoundError as e:
            # A ModuleNotFoundError is raised if the plugin is not installed
            # or if the plugin itself has issues importing code. Distinguish
            # between the two.
            if pyimport in e.args[0]:
                raise qik.errors.PluginNotFound(
                    f"Plugin '{pyimport}' could not be imported. "
                    f"Make sure it's installed and has a 'qikplugin' module."
                ) from e
            else:
                # TODO: Show that this is an unexpected error. Currently CLI
                # users get a raw stack trace.
                raise


class BaseDynamicPlugins(Base, frozen=True):
    def items(self) -> Generator[tuple[str, Any]]:  # type: ignore
        for key in self.__struct_fields__:
            yield (key, getattr(self, key))


def _parse_project_config(contents: bytes, plugins_conf: Plugins) -> Project:
    """Dynamically generate a Project config class based on installed plugins."""

    dep_plugin_types = _PLUGIN_TYPES["dep"]
    cache_plugin_types = _PLUGIN_TYPES["cache"]
    venv_plugin_types = _PLUGIN_TYPES["venv"]

    DynamicCacheTypes = Union[(Cache, *(cls for cls, _ in cache_plugin_types.values()))]
    DynamicVenvTypes = Union[
        (ActiveVenv, SpaceVenv, *(cls for cls, _ in venv_plugin_types.values()))
    ]
    DynamicDeps = Union[
        (
            str,
            GlobDep,
            CmdDep,
            PydistDep,
            ConstDep,
            LoadDep,
            *(cls for cls, _ in dep_plugin_types.values()),
        )
    ]
    DynamicBaseConf = msgspec.defstruct(
        "DynamicBaseConf",
        [("deps", list[DynamicDeps], [])],  # type: ignore
        bases=(BaseConf,),
        frozen=True,
    )

    DynamicSpace = msgspec.defstruct(
        "DynamicSpace",
        [("venv", DynamicVenvTypes | str | list[str] | None, None)],  # type: ignore
        bases=(Space,),
        frozen=True,
    )
    DynamicCmd = msgspec.defstruct(
        "DynamicCmd", [("deps", list[DynamicDeps], [])], bases=(Cmd,), frozen=True
    )

    existing_plugin_confs = dict.fromkeys(plugins_conf.plugins_by_pyimport, PluginLocator)
    DynamicPlugins = msgspec.defstruct(
        "DynamicPlugins",
        [
            (
                plugins_conf.plugins_by_pyimport[plugin_pyimport][0],
                plugin_conf | str,  # type: ignore
            )
            for plugin_pyimport, plugin_conf in (existing_plugin_confs | _CONF_TYPES).items()
        ],
        bases=(BaseDynamicPlugins,),
        frozen=True,
    )

    DynamicDefaults = msgspec.defstruct(
        "DynamicDefaults",
        [("venv", DynamicVenvTypes | str | list[str] | None, None)],  # type: ignore
        bases=(Defaults,),
        frozen=True,
    )

    # Must register as part of the global namespace in order for msgspec to
    # recognize dynamic nested type.
    globals()["DynamicSpace"] = DynamicSpace
    globals()["DynamicCmd"] = DynamicCmd
    globals()["DynamicCacheTypes"] = DynamicCacheTypes
    globals()["DynamicDeps"] = DynamicDeps
    globals()["DynamicBaseConf"] = DynamicBaseConf
    globals()["DynamicPlugins"] = DynamicPlugins
    globals()["DynamicDefaults"] = DynamicDefaults

    DynamicProject = msgspec.defstruct(
        "DynamicProject",
        [
            ("venvs", dict[str, DynamicVenvTypes], {}),
            ("caches", dict[str, DynamicCacheTypes], {}),
            ("spaces", dict[str, DynamicSpace | str | list[str]], {}),
            ("commands", dict[str, DynamicCmd | str], {}),
            ("base", DynamicBaseConf, msgspec.field(default_factory=DynamicBaseConf)),
            ("defaults", DynamicDefaults, msgspec.field(default_factory=DynamicDefaults)),
            ("plugins", DynamicPlugins, msgspec.field(default_factory=DynamicPlugins)),
        ],
        bases=(Project,),
        frozen=True,
    )

    return msgspec.toml.decode(contents, type=DynamicProject)  # type: ignore


@qik.func.cache
def load() -> tuple[Project, pathlib.Path]:
    """Load the project configuration and file.

    This serves as an entry point for all of qik, so we set the python path
    here too.
    """
    cwd = pathlib.Path.cwd()
    qik_toml: pathlib.Path | None = None

    for directory in (cwd, *cwd.parents):
        if (directory / "qik.toml").is_file():
            qik_toml = directory / "qik.toml"

        if (directory / ".git").is_dir():
            break

    if qik_toml:
        try:
            contents = qik_toml.read_bytes()
            plugins_conf = msgspec.toml.decode(contents, type=Plugins)
            _load_plugins(plugins_conf)
            conf = _parse_project_config(contents, plugins_conf)
        except msgspec.ValidationError as e:
            raise qik.errors.ConfigParse(f"Error parsing qik.toml: {e}") from e

        python_path = qik_toml.parent / conf.defaults.python_path
        sys.path.insert(0, str(python_path))
        return conf, qik_toml
    else:
        raise qik.errors.ConfigNotFound("Could not locate qik.toml configuration file.")


@qik.func.cache
def project() -> Project:
    return load()[0]


@qik.func.cache
def module_locator(uri: str, *, by_path: bool = False) -> ModuleLocator:
    """Get module locator."""
    proj = project()
    lookup = proj.modules_by_path if by_path else proj.modules_by_name
    if uri not in lookup:
        raise qik.errors.ModuleNotFound(f'Module "{uri}" not configured in {location().name}.')

    return lookup[uri]


@qik.func.cache
def plugin_locator(uri: str, *, by_pyimport: bool = False) -> tuple[str, PluginLocator]:
    """Get plugin locator."""
    proj = project()
    lookup = proj.plugins_by_pyimport if by_pyimport else proj.plugins_by_name
    if uri not in lookup:
        raise qik.errors.PluginNotFound(f'Plugin "{uri}" not configured in {location().name}.')

    if by_pyimport:
        return proj.plugins_by_pyimport[uri]
    else:
        return uri, proj.plugins_by_name[uri]


@qik.func.cache
def plugin(pyimport: str) -> PluginConf:
    """Get dynamic plugin configuration."""
    proj = project()
    conf = getattr(proj.plugins, proj.plugins_by_pyimport[pyimport][0])
    if isinstance(conf, str):
        return _CONF_TYPES[pyimport]()  # type: ignore
    else:
        return conf  # type: ignore


@qik.func.cache
def search(name: str | None = None) -> ModuleOrPlugin:
    """Search configuration for a given module, plugin, or project."""
    if not name:
        return project()
    else:
        try:
            return module_locator(name).conf
        except qik.errors.ModuleNotFound:
            try:
                return plugin_locator(name)[1].conf
            except qik.errors.PluginNotFound:
                raise qik.errors.ModuleOrPluginNotFound(
                    f'Module or plugin "{name}" not configured in {location().name}.'
                ) from None


@qik.func.cache
def uri_parts(uri: str) -> tuple[str | None, str]:
    """Return the module and name of a URI."""
    if "/" in uri:
        return tuple(uri.rsplit("/", 1))  # type: ignore
    elif "." in uri:
        return tuple(uri.rsplit(".", 1))  # type: ignore
    else:
        return (None, uri)


@qik.func.cache
def command(uri: str) -> Cmd:
    """Get configuration for a command."""
    module, name = uri_parts(uri)
    conf = search(module)
    if name not in conf.commands:
        raise qik.errors.CommandNotFound(f'Command "{uri}" not configured.')

    cmd_conf = conf.commands[name]
    return Cmd(exec=cmd_conf) if isinstance(cmd_conf, str) else cmd_conf


@qik.func.cache
def root() -> pathlib.Path:
    """Get the absolute root project directory."""
    return load()[1].parent


@qik.func.cache
def abs_python_path() -> pathlib.Path:
    """Get the absolute python path."""
    return load()[1].parent / project().defaults.python_path


@qik.func.cache
def priv_work_dir(rel: bool = False) -> pathlib.Path:
    """Get the private work directory."""
    return root() / "._qik" if not rel else pathlib.Path("._qik")


@qik.func.cache
def pub_work_dir(rel: bool = False) -> pathlib.Path:
    """Get the public work directory."""
    return root() / ".qik" if not rel else pathlib.Path(".qik")


@qik.func.cache
def location() -> pathlib.Path:
    """Get the root configuration file."""
    return load()[1]


@qik.func.cache
def pyimport(path: str) -> str:
    """Return the python import for a path."""
    python_path = project().defaults.python_path
    if python_path == ".":
        return path.replace("/", ".")
    else:
        return str(pathlib.Path(path).relative_to(python_path)).replace("/", ".")


def default_venv_type() -> type[Venv]:
    """Get the default venv type."""
    load()
    return _VENV_TYPE or ActiveVenv



from __future__ import annotations

import argparse
import os
import pathlib
import sys

import qik.conf
import qik.ctx
import qik.errors
import qik.runner
import qik.space
import qik.unset


def _get_working_space() -> str | None:
    if pathlib.Path.cwd() != qik.conf.root():
        location = (
            str(pathlib.Path.cwd().relative_to(qik.conf.root())).replace(os.path.sep, "/") + "/"
        )
        for space_name, space_conf in qik.conf.project().resolved_spaces.items():
            if space_conf.root and location.startswith(space_conf.root):
                return space_name


@qik.errors.catch_and_exit()
def qik_entry() -> None:
    """The entrypoint into the qik CLI."""
    parser = argparse.ArgumentParser()

    parser.add_argument("commands", help="Command name(s)", nargs="*")
    parser.add_argument(
        "-m",
        "--module",
        help="Select commands by module(s).",
        action="append",
        dest="modules",
    )
    parser.add_argument(
        "-s",
        "--space",
        help="Select commands by space(s).",
        action="append",
        dest="spaces",
    )
    parser.add_argument(
        "-t",
        "--tag",
        help="Select by tag(s).",
        action="append",
        dest="tags",
    )
    parser.add_argument(
        "--watch", action="store_true", help="Watch for changes.", default=qik.unset.UNSET
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Don't read caches.",
        default=qik.unset.UNSET,
    )
    parser.add_argument(
        "-n", "--workers", help="Number of workers.", type=int, default=qik.unset.UNSET
    )
    parser.add_argument(
        "--isolated",
        action="store_true",
        help="Don't run dependent commands.",
        default=qik.unset.UNSET,
    )
    parser.add_argument(
        "--ls",
        dest="ls",
        action="store_true",
        help="List selected commands.",
        default=qik.unset.UNSET,
    )
    parser.add_argument(
        "--fail",
        action="store_true",
        help="Fail if any commands are selected.",
        default=qik.unset.UNSET,
    )
    parser.add_argument("--since", help="Select since git SHA.", default=qik.unset.UNSET)
    parser.add_argument(
        "--cache",
        help="Select by cache(s).",
        action="append",
        dest="caches",
    )
    parser.add_argument(
        "--cache-status",
        help="Select by cache status.",
        default=qik.unset.UNSET,
        choices=["warm", "cold"],
    )
    parser.add_argument(
        "-l",
        "--logger",
        help="Set the logger.",
        default=qik.unset.UNSET,
        choices=["progress", "stdout"],
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        nargs="?",
        const=2,
        default=qik.unset.UNSET,
        type=int,
        help="Set verbosity (1 by default, 2 if -v is present, or specify level)",
    )

    args = parser.parse_args()
    spaces = args.spaces

    # Set the space if in a space root
    if not spaces:
        if working_space := _get_working_space():
            spaces = [working_space]

    with qik.ctx.set_vars(
        "qik",
        watch=args.watch,
        force=args.force,
        isolated=args.isolated,
        ls=args.ls,
        workers=args.workers,
        since=args.since,
        fail=args.fail,
        cache_status=args.cache_status,
        verbosity=args.verbosity,
        commands=args.commands or qik.unset.UNSET,
        modules=args.modules or qik.unset.UNSET,
        spaces=spaces or qik.unset.UNSET,
        caches=args.caches or qik.unset.UNSET,
        tags=args.tags or qik.unset.UNSET,
        logger=args.logger,
    ):
        res = qik.runner.exec()
        qik_ctx = qik.ctx.by_namespace("qik")

        if qik_ctx.ls:
            for cmd in sorted(res, key=lambda k: k.name):
                print(cmd.name)

        if res and qik_ctx.fail:
            sys.exit(1)


@qik.errors.catch_and_exit()
def qikx_entry() -> None:
    """The entrypoint into the qikx CLI."""
    if len(sys.argv) < 2 or sys.argv[1] == "--help":
        print("Usage: qikx <command@space command_args>")
        sys.exit(1)

    space = None
    command = sys.argv[1]
    if command in ("--install", "-i"):
        venv_type = qik.conf.default_venv_type()
        install_cmd = venv_type.install_cmd
        if install_cmd:
            os.execvp("qik", ["qik", install_cmd])
    else:
        args = sys.argv[2:]

        if "@" in command:
            command, space = command.split("@", 1)

        # Set the space if in a space root
        if not space:
            space = _get_working_space() or "default"

        resolved_space = qik.space.load(space)
        os.environ |= resolved_space.environ
        os.execvp(command, [command, *args])



from __future__ import annotations

import qik.conf
import qik.ctx
import qik.func
import qik.unset


class PygraphDepConf(qik.conf.Dep, tag="pygraph", frozen=True):
    pyimport: str


class PygraphPluginConf(qik.conf.PluginConf, frozen=True, dict=True, tag="qik.pygraph"):
    ignore_type_checking: bool = False
    ignore_pydists: bool = False
    cache: str | qik.unset.UnsetType = qik.unset.UNSET
    build_cache: str | qik.unset.UnsetType = qik.unset.UNSET
    lock_cache: str | qik.unset.UnsetType = qik.unset.UNSET
    check_cache: str | qik.unset.UnsetType = qik.unset.UNSET

    @qik.func.cached_property
    def resolved_build_cache(self) -> str:
        return qik.ctx.format(
            qik.unset.coalesce(
                self.build_cache,
                self.cache,
                qik.conf.project().defaults.cache,
                default="local",
            )
        )

    @qik.func.cached_property
    def resolved_lock_cache(self) -> str:
        return qik.ctx.format(
            qik.unset.coalesce(
                self.lock_cache,
                self.cache,
                qik.conf.project().defaults.cache,
                default="local",
            )
        )

    @qik.func.cached_property
    def resolved_check_cache(self) -> str:
        return qik.ctx.format(
            qik.unset.coalesce(
                self.check_cache,
                self.cache,
                qik.conf.project().defaults.cache,
                default="local",
            )
        )


qik.conf.register_type(PygraphDepConf, "qik.pygraph.dep.factory")
qik.conf.register_conf(PygraphPluginConf)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

import qik.conf
import qik.ctx
import qik.dep
import qik.pygraph.utils

if TYPE_CHECKING:
    from qik.pygraph.qikplugin import PygraphDepConf


class PygraphDep(qik.dep.BaseCmd, frozen=True):
    """A python module and its associated imports."""

    strict: ClassVar[bool] = True  # type: ignore
    space: str | None = None

    def get_cmd_name(self) -> str:
        return qik.pygraph.utils.lock_cmd_name()

    def get_cmd_args(self) -> dict[str, str | None]:
        return {"pyimport": self.val, "space": self.space}

    @property
    def globs(self) -> list[str]:  # type: ignore
        return [str(qik.pygraph.utils.lock_path(self.val, self.space))]


def factory(
    conf: PygraphDepConf,
    module: qik.conf.ModuleLocator | None = None,
    space: str | None = None,
) -> PygraphDep:
    return PygraphDep(qik.ctx.format(conf.pyimport, module=module), space=space)



import pkgutil
from typing import Generator, Iterator

import msgspec

import qik.conf
import qik.errors
import qik.func
import qik.unset
import qik.venv


@qik.func.per_run_cache
def _read_dotenv(path: str) -> dict[str, str]:
    def _iter_dotenv_lines() -> Generator[tuple[str, str], None, None]:
        with open(path, "r") as file:
            for line in file:
                line = line.strip()
                if line and not line.startswith("#"):
                    key, value = line.split("=", 1)
                    yield key, value

    try:
        return dict(_iter_dotenv_lines())
    except FileNotFoundError as e:
        raise qik.errors.DotEnvNotFound(f'Dotenv file "{path}" not found.') from e


class Space(msgspec.Struct, frozen=True, dict=True):
    name: str
    conf: qik.conf.Space

    def _fence_pyimports_iter(self) -> Iterator[str]:
        for locator in self.conf.modules_by_path.values():
            yield locator.pyimport

        if isinstance(self.conf.fence, list):
            for value in self.conf.fence:
                if isinstance(value, str):
                    yield qik.conf.pyimport(value)
                else:
                    space = load(value.name)
                    yield from space._fence_pyimports_iter()

    def _fence_paths_iter(self) -> Iterator[str]:
        for locator in self.conf.modules_by_path.values():
            yield locator.path

        if isinstance(self.conf.fence, list):
            for value in self.conf.fence:
                if isinstance(value, str):
                    yield value
                elif value.name != self.name:
                    space = load(value.name)
                    yield from space._fence_paths_iter()

    @qik.func.cached_property
    def fence_pyimports(self) -> list[str]:
        try:
            return sorted(set(self._fence_pyimports_iter()))
        except RecursionError as e:
            raise qik.errors.CircularFence("Circular fence detected.") from e

    @qik.func.cached_property
    def fence_paths(self) -> list[str]:
        try:
            return sorted(set(self._fence_paths_iter()))
        except RecursionError as e:
            raise qik.errors.CircularFence("Circular fence detected.") from e

    @qik.func.cached_property
    def venv(self) -> qik.venv.Venv:
        conf = qik.unset.coalesce(self.conf.venv, qik.conf.project().defaults.venv, default=None)

        if isinstance(conf, str | list):
            conf = qik.conf.default_venv_type()(reqs=conf)

        try:
            if conf is None or isinstance(conf, qik.conf.ActiveVenv):
                return qik.venv.active()
            elif isinstance(conf, qik.conf.SpaceVenv):
                if conf.name == self.name:
                    return qik.venv.active()
                else:
                    return load(conf.name).venv
            else:
                factory = qik.conf.get_type_factory(conf)
                return pkgutil.resolve_name(factory)(self.name, conf)
        except RecursionError as e:
            raise qik.errors.CircularVenv("Circular venv detected.") from e

    @property
    def dotenvs(self) -> list[str]:
        dotenv_val = qik.unset.coalesce(
            self.conf.dotenv, qik.conf.project().defaults.dotenv, default=[]
        )
        return [dotenv_val] if isinstance(dotenv_val, str) else dotenv_val

    @qik.func.cached_property
    def environ(self) -> dict[str, str]:
        env = self.venv.environ
        for dotenv in self.dotenvs:
            env.update(_read_dotenv(dotenv))

        return env

    @qik.func.cached_property
    def glob_deps(self) -> set[str]:
        return set(self.dotenvs)


@qik.func.cache
def load(name: str = "default") -> Space:
    """Get configuration for a space."""
    proj = qik.conf.project()
    if name != "default" and name not in proj.resolved_spaces:
        raise qik.errors.SpaceNotFound(f'Space "{name}" not configured.')

    return Space(name=name, conf=proj.resolved_spaces.get(name, qik.conf.Space()))



import qik.conf
import qik.func


@qik.func.cache
def lock_cmd_name() -> str:
    plugin_name = qik.conf.plugin_locator("qik.uv", by_pyimport=True)[0]
    return f"{plugin_name}.lock"


@qik.func.cache
def install_cmd_name() -> str:
    plugin_name = qik.conf.plugin_locator("qik.uv", by_pyimport=True)[0]
    return f"{plugin_name}.install"



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any, Iterator, Literal, TypeAlias

import qik.func
import qik.lazy

if TYPE_CHECKING:
    import rich.console as rich_console

    Emoji: TypeAlias = Literal[
        "broken_heart",
        "construction",
        "eyes",
        "white_check_mark",
        "fast-forward_button",
        "heavy_minus_sign",
    ]
    Color: TypeAlias = Literal["cyan", "red", "green", "yellow", "white"]
else:
    rich_console = qik.lazy.module("rich.console")


def fmt_msg(msg: str, emoji: Emoji | None = None, color: Color | None = None) -> str:
    if color:
        msg = f"[{color}]{msg}[/{color}]"

    if emoji:
        msg = f":{emoji}-emoji: {msg}"

    return msg


def print(msg: str, emoji: Emoji | None = None, color: Color | None = None, **kwargs: Any) -> None:
    get().print(fmt_msg(msg, emoji=emoji, color=color), highlight=False, **kwargs)


def print_exception() -> None:
    get().print_exception()


def rule(
    msg: str = "", emoji: Emoji | None = None, color: Color | None = None, **kwargs: Any
) -> None:
    get().rule(
        fmt_msg(msg, emoji=emoji, color=color), align="left", style=color or "rule.line", **kwargs
    )


@contextlib.contextmanager
def capture() -> Iterator[None]:
    with get().capture():
        yield


@qik.func.cache
def get() -> rich_console.Console:
    return qik.lazy.object(rich_console.Console)  # type: ignore



from __future__ import annotations

import concurrent.futures
import os
import pathlib
from typing import TYPE_CHECKING

import boto3.s3
import msgspec

import qik.cache
import qik.ctx
import qik.file
import qik.func

if TYPE_CHECKING:
    import boto3
    from boto3.resources.base import ServiceResource

    from qik.runnable import Runnable
    from qik.s3.qikplugin import S3Conf
else:
    import qik.lazy

    boto3 = qik.lazy.module("boto3")


def _download_file(*, bucket, obj, dir: pathlib.Path, prefix: pathlib.Path):
    target = dir / os.path.relpath(obj.key, str(prefix))

    if obj.key[-1] == "/":
        return  # Skip directories

    try:
        bucket.download_file(obj.key, target)
    except FileNotFoundError:
        qik.file.make_parent_dirs(target)
        bucket.download_file(obj.key, target)

    return obj.key


def _upload_file(*, bucket, prefix: pathlib.Path, path: pathlib.Path, dir: pathlib.Path) -> str:
    relative_path = os.path.relpath(path, dir)
    s3_key = str((prefix / relative_path)).replace("\\", "/")
    bucket.upload_file(path, s3_key)
    return s3_key


class Client(msgspec.Struct, frozen=True, dict=True):
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    aws_session_token: str | None = None
    region_name: str | None = None
    endpoint_url: str | None = None

    @qik.func.cached_property
    def s3_session(self) -> ServiceResource:
        s3_kwargs = {"endpoint_url": self.endpoint_url} if self.endpoint_url else {}
        return boto3.Session(  # type: ignore
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            region_name=self.region_name,
        ).resource("s3", **s3_kwargs)  # type: ignore

    def download_dir(
        self, *, bucket_name: str, prefix: pathlib.Path, dir: pathlib.Path, max_workers: int = 10
    ) -> None:
        bucket = self.s3_session.Bucket(bucket_name)  # type: ignore

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [
                executor.submit(_download_file, bucket=bucket, obj=obj, dir=dir, prefix=prefix)
                for obj in bucket.objects.filter(Prefix=str(prefix))
            ]

            for future in concurrent.futures.as_completed(futures):
                # TODO: Better handle partial upload failures
                future.result()

    def upload_dir(
        self, *, bucket_name: str, prefix: pathlib.Path, dir: pathlib.Path, max_workers: int = 10
    ) -> None:
        bucket = self.s3_session.Bucket(bucket_name)  # type: ignore

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [
                executor.submit(
                    _upload_file,
                    prefix=prefix,
                    bucket=bucket,
                    path=pathlib.Path(root) / file,
                    dir=dir,
                )
                for root, _, files in os.walk(dir)
                for file in files
            ]

            for future in concurrent.futures.as_completed(futures):
                # TODO: Better handle partial upload failures
                future.result()


class S3Cache(msgspec.Struct, qik.cache.Local, frozen=True, dict=True):
    """A custom cache using the S3 backend"""

    bucket: str
    prefix: str = ""
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    aws_session_token: str | None = None
    region_name: str | None = None
    endpoint_url: str | None = None

    def remote_path(self, *, runnable: Runnable, hash: str) -> pathlib.Path:
        return pathlib.Path(self.prefix) / f"{runnable.slug}-{hash}"

    @qik.func.cached_property
    def client(self) -> Client:
        return Client(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            region_name=self.region_name,
            endpoint_url=self.endpoint_url,
        )

    def on_miss(self, *, runnable: Runnable, hash: str) -> None:
        super().pre_get(runnable=runnable, hash=hash)

        self.client.download_dir(
            bucket_name=self.bucket,
            prefix=self.remote_path(runnable=runnable, hash=hash),
            dir=self.base_path(runnable=runnable, hash=hash),
        )

    def post_set(self, *, runnable: Runnable, hash: str, manifest: qik.cache.Manifest) -> None:
        super().post_set(runnable=runnable, hash=hash, manifest=manifest)

        self.client.upload_dir(
            bucket_name=self.bucket,
            prefix=self.remote_path(runnable=runnable, hash=hash),
            dir=self.base_path(runnable=runnable, hash=hash),
        )


def factory(name: str, conf: S3Conf) -> S3Cache:
    endpoint_url = qik.ctx.format(conf.endpoint_url)
    endpoint_url = None if endpoint_url == "None" else endpoint_url
    return S3Cache(
        bucket=qik.ctx.format(conf.bucket),
        prefix=qik.ctx.format(conf.prefix),
        aws_access_key_id=qik.ctx.format(conf.aws_access_key_id),
        aws_secret_access_key=qik.ctx.format(conf.aws_secret_access_key),
        aws_session_token=qik.ctx.format(conf.aws_session_token),
        region_name=qik.ctx.format(conf.region_name),
        endpoint_url=endpoint_url,
    )



from __future__ import annotations

import pathlib
from typing import cast

import qik.conf
import qik.dep
import qik.errors
import qik.func
import qik.runnable
import qik.space
import qik.uv.conf
import qik.uv.utils
import qik.uv.venv
import qik.venv


def lock_cmd_factory(
    cmd: str, conf: qik.conf.Cmd, **args: str
) -> dict[str, qik.runnable.Runnable]:
    space = args.get("space")
    if not space:
        raise qik.errors.ArgNotSupplied('"space" arg is required for qik.uv.lock command.')

    venv = cast(qik.uv.venv.UVVenv, qik.space.load(space).venv)
    cmd_name = qik.uv.utils.lock_cmd_name()
    uv_conf = qik.uv.conf.get()

    pip_compile = f"uv pip compile --universal {' '.join(venv.reqs)} -o {venv.lock}"
    constraint_deps = []
    if venv.constraint:
        pip_compile = f"{pip_compile} --constraint {venv.constraint}"
        constraint_deps = [qik.dep.Glob(venv.constraint)]

    environ = {"UV_CACHE_DIR": str(qik.conf.priv_work_dir(rel=True) / ".uv")}

    if uv_conf.resolved_index_url:
        environ["UV_INDEX_URL"] = uv_conf.resolved_index_url

    if uv_conf.resolved_extra_index_url:
        environ["UV_EXTRA_INDEX_URL"] = uv_conf.resolved_extra_index_url

    runnable = qik.runnable.Runnable(
        name=qik.runnable.fmt_name(cmd_name, space=space),
        cmd=cmd_name,
        val=f"mkdir -p {pathlib.Path(venv.lock).parent} && {pip_compile}",
        deps=[*qik.dep.base(), *(qik.dep.Glob(req) for req in venv.reqs), *constraint_deps],
        artifacts=[venv.lock],
        cache=uv_conf.resolved_cache,
        cache_when="success",
        space=space,
        venv=None,
        environ=environ,
    )
    return {runnable.name: runnable}


def install_cmd_factory(
    cmd: str, conf: qik.conf.Cmd, **args: str
) -> dict[str, qik.runnable.Runnable]:
    space = args.get("space")
    if not space:
        raise qik.errors.ArgNotSupplied('"space" arg is required for qik.uv.install command.')

    venv = cast(qik.uv.venv.UVVenv, qik.space.load(space).venv)
    venv_python = f"--python '{venv.python}'" if venv.python else ""
    cmd_name = qik.uv.utils.install_cmd_name()
    runnable = qik.runnable.Runnable(
        name=qik.runnable.fmt_name(cmd_name, space=space),
        cmd=cmd_name,
        val=f"uv venv {venv.dir.relative_to(qik.conf.root())} {venv_python} && uv pip sync {venv.lock} --python {venv.dir}/bin/python",
        deps=[
            *qik.dep.base(),
            qik.dep.Cmd(qik.uv.utils.lock_cmd_name(), args={"space": space}, strict=True),
            qik.dep.Glob(venv.lock),
        ],
        artifacts=[],
        cache="local",
        cache_when="success",
        space=space,
        venv=None,
    )
    return {runnable.name: runnable}



"""Functions and classes for managing and serializing the import graph for a module."""

from __future__ import annotations

import importlib.machinery
import importlib.util
import pkgutil
import sys
from typing import TYPE_CHECKING, Iterator, Literal, overload

import msgspec
from typing_extensions import Self

import qik.conf
import qik.ctx
import qik.dep
import qik.func
import qik.pygraph.conf
import qik.shell

if TYPE_CHECKING:
    import grimp
    import rustworkx as rx
else:
    import qik.lazy

    grimp = qik.lazy.module("grimp")
    rx = qik.lazy.module("rustworkx")


class Module(msgspec.Struct, array_like=True, omit_defaults=True, frozen=True, dict=True):
    imp: str
    is_internal: bool = True

    @qik.func.cached_property
    def path(self) -> str:
        return self.imp.replace(".", "/")


class Graph(msgspec.Struct, frozen=True, dict=True):
    modules: list[Module]
    edges: list[tuple[int, int]]

    def bind(self, rx_graph: rx.PyDiGraph) -> Self:
        self.__dict__["_rx"] = rx_graph
        return self

    @property
    def rx(self) -> rx.PyDiGraph:
        if "_rx" not in self.__dict__:
            g = rx.PyDiGraph()
            g.add_nodes_from(self.modules)
            g.add_edges_from_no_data(self.edges)
            self.__dict__["_rx"] = g

        return self.__dict__["_rx"]

    @qik.func.cached_property
    def modules_idx(self) -> dict[str, int]:
        return {module.imp: i for i, module in enumerate(self.modules)}

    @overload
    def upstream_imports(self, imp: str, /, *, idx: Literal[True]) -> list[tuple[int, int]]: ...

    @overload
    def upstream_imports(
        self, imp: str, /, *, idx: Literal[False]
    ) -> list[tuple[Module, Module]]: ...

    @overload
    def upstream_imports(self, imp: str, /) -> list[tuple[Module, Module]]: ...

    @overload
    def upstream_imports(
        self, imp: str, /, *, idx: bool = ...
    ) -> list[tuple[Module, Module]] | list[tuple[int, int]]: ...

    def upstream_imports(
        self, imp: str, /, *, idx: bool = False
    ) -> list[tuple[Module, Module]] | list[tuple[int, int]]:
        upstream = rx.digraph_dfs_edges(self.rx, self.modules_idx[imp])
        if idx:
            return list(upstream)
        else:
            modules = self.modules
            return [(modules[edge[0]], modules[edge[1]]) for edge in upstream]

    @overload
    def upstream_modules(self, imp: str, /, *, idx: Literal[True]) -> set[int]: ...

    @overload
    def upstream_modules(self, imp: str, /, *, idx: Literal[False]) -> set[Module]: ...

    def upstream_modules(self, imp: str, /, *, idx: bool = False) -> set[Module] | set[int]:
        return {imp[1] for imp in self.upstream_imports(imp, idx=idx)}  # type: ignore


def build() -> Graph:
    """Build the import graph from the current codebase."""
    python_path = str(qik.conf.abs_python_path())
    internal_modules = {
        module.name
        for module in pkgutil.iter_modules()
        if module.ispkg
        and isinstance(module.module_finder, importlib.machinery.FileFinder)
        and module.module_finder.path == python_path
    }
    # TODO: Raise a runtime error if no internalmodules were found. Suggest to configure the python path.
    stdlib_modules = sys.stdlib_module_names | set(sys.builtin_module_names)

    # Parse the codebase with grimp
    pygraph_conf = qik.pygraph.conf.get()
    grimp_g = grimp.build_graph(
        *internal_modules,
        include_external_packages=not pygraph_conf.ignore_pydists,
        exclude_type_checking_imports=pygraph_conf.ignore_type_checking,
        cache_dir=str(qik.conf.priv_work_dir() / ".grimp"),
    )
    graph_modules_imps = [(imp.split(".", 1)[0], imp) for imp in sorted(grimp_g.modules)]
    modules = [
        Module(imp=imp, is_internal=module in internal_modules)
        for module, imp in graph_modules_imps
        if module not in stdlib_modules
    ]
    modules_idx = {module.imp: i for i, module in enumerate(modules)}

    # Start constructing the rustworkx graph
    rx_g = rx.PyDiGraph()
    rx_g.add_nodes_from(modules)

    # Add layered modules as dependencies on one another for graph locking
    def _iter_layered_module_edges() -> Iterator[tuple[int, int]]:
        for i, module in enumerate(modules):
            if i > 0 and "." in module.imp:
                for parent in modules[i - 1 :: -1]:
                    if found := module.imp.startswith(f"{parent.imp}."):
                        yield (modules_idx[parent.imp], i)

                    if found or "." not in parent.imp:
                        break

    rx_g.add_edges_from_no_data(list(_iter_layered_module_edges()))

    # Add edges for imports
    for module in modules:
        module_idx = modules_idx[module.imp]
        rx_g.add_edges_from_no_data(
            [
                (module_idx, modules_idx[imported])
                for imported in grimp_g.find_modules_directly_imported_by(module.imp)
                if imported in modules_idx
            ]
        )

    return Graph(modules=modules, edges=sorted(rx_g.edge_list())).bind(rx_g)



"""Custom caching decorators similar to functools."""

from __future__ import annotations

import functools
from typing import Callable, Final, ParamSpec, TypeVar

P = ParamSpec("P")
T = TypeVar("T")


_RUN_CACHED_FUNCS: Final[list[Callable]] = []
_RUN_CACHED_PROPS: Final[list[tuple[dict, str]]] = []


cache = functools.cache
lru_cache = functools.lru_cache
cached_property = functools.cached_property


def per_run_cache(func: Callable[P, T]) -> Callable[P, T]:
    """Cache a function's results for the duration of a run."""

    cached_func = cache(func)
    _RUN_CACHED_FUNCS.append(cached_func)
    return cached_func  # type: ignore


_NOT_FOUND = object()


class per_run_cached_property(functools.cached_property):
    def __get__(self, instance, owner=None):
        try:
            val = instance.__dict__.get(self.attrname, _NOT_FOUND)
            if val is _NOT_FOUND:
                val = self.func(instance)
                instance.__dict__[self.attrname] = val
                _RUN_CACHED_PROPS.append((instance.__dict__, self.attrname))  # type: ignore

            return val
        except AttributeError:
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None


def clear_per_run_cache() -> None:
    """Clear all run cache."""
    for cache in _RUN_CACHED_FUNCS:
        cache.cache_clear()  # type: ignore

    for dictionary, attr in _RUN_CACHED_PROPS:
        if attr in dictionary:
            del dictionary[attr]



from __future__ import annotations

import collections
import concurrent.futures
import copy
import pathlib
import sys
from typing import TYPE_CHECKING, Iterable, Iterator, Literal, TypeAlias

from typing_extensions import Self

import qik.cmd
import qik.conf
import qik.console
import qik.ctx
import qik.dep
import qik.errors
import qik.func
import qik.logger
import qik.runnable
import qik.shell
import qik.unset
import qik.watcher

if TYPE_CHECKING:
    from qik.cache import Entry
    from qik.runnable import Result, Runnable

    Direction: TypeAlias = Literal["up", "down"]
    Edges: TypeAlias = dict[str, set[str]]


class DAGPool:
    def __init__(self, graph: Graph):
        self.graph = graph
        self.nodes = graph.nodes
        self.upstream = graph.upstream

        runner = qik.ctx.runner()
        self.pool = runner.pool
        self.logger = runner.logger
        self.downstream = collections.defaultdict(set)
        for node, deps in graph.upstream.items():
            for dep in deps:
                self.downstream[dep].add(node)

    def _exec(self) -> dict[str, Result | None]:
        in_degree = {name: len(deps) for name, deps in self.upstream.items()}
        results: dict[str, Result | None] = {}
        failed: set[str] = set()
        futures: dict[str, concurrent.futures.Future] = {}
        exception: Exception | None = None

        def _skip(name: str):
            if name in in_degree:
                del in_degree[name]

            failed.add(name)
            results[name] = None
            self.logger.print(
                msg=f"{name} [default][dim]{self.nodes[name].val}",
                emoji="heavy_minus_sign",
                color="yellow",
                runnable=self.nodes[name],
                event="finish",
                result=None,
            )

        def _finish(future: concurrent.futures.Future) -> Exception | None:
            name = next(name for name, val in futures.items() if val == future)
            exception: Exception | None = None
            try:
                result = future.result()
            except Exception as exc:
                result = qik.runnable.Result(log=str(exc), code=1, hash="")
                exception = exc

            results[name] = result
            if result.code != 0:
                failed.add(name)
                for dep in self.downstream[name]:
                    _skip(dep)
            else:
                for dep in self.downstream[name]:
                    if dep in in_degree:
                        in_degree[dep] -= 1

            del in_degree[name]
            del futures[name]

            return exception

        while in_degree:
            ready_tasks = (name for name, degree in in_degree.items() if degree == 0)
            for name in ready_tasks:
                if name not in futures:
                    if not any(dep in failed for dep in self.upstream[name]):
                        futures[name] = self.pool.submit(self.nodes[name].exec)
                    else:
                        _skip(name)

            finished, unfinished = concurrent.futures.wait(
                futures.values(), return_when=concurrent.futures.FIRST_COMPLETED
            )
            if not finished and not unfinished:
                break

            for future in finished:
                exception = _finish(future) or exception

        for name, future in futures.items():
            future.cancel()
            results[name] = None

        if exception:
            raise exception

        return results

    def exec(self) -> dict[str, Result | None]:
        if not self.nodes:
            return {}
        else:
            with self.logger.run(self.graph):
                return self._exec()


class Graph:
    """A graph of runnables."""

    def __init__(
        self,
        cmds: Iterable[str],
        *,
        modules: Iterable[str] | None = None,
        spaces: Iterable[str] | None = None,
    ):
        self.cmds = cmds
        self._view = None
        modules = frozenset(modules or [])
        spaces = frozenset(spaces or [])

        # Construct the graph
        self._nodes: dict[str, Runnable] = {
            runnable.name: runnable
            for cmd in cmds
            for runnable in qik.cmd.runnables(cmd)
            if (
                (not modules or runnable.module in modules)
                and (not spaces or runnable.space in spaces)
            )
        }
        self._graph = {
            "up": collections.defaultdict(set[str]),
            "down": collections.defaultdict(set[str]),
        }

        def _runnable_edges(runnable: Runnable) -> Iterator[tuple[Runnable, Runnable, Direction]]:
            """Given a runnable, iterate over all of its edges."""
            for dep in runnable.deps_collection.runnables.values():
                isolated = (
                    dep.isolated
                    if not isinstance(dep.isolated, qik.unset.UnsetType)
                    else qik.ctx.by_namespace("qik").isolated
                )
                if not isolated or dep.obj.name in self._nodes:
                    yield (runnable, dep.obj, "up")
                    if dep.strict:
                        yield (dep.obj, runnable, "down")

                    yield from _runnable_edges(dep.obj)

        def _runnables_edges() -> Iterator[tuple[Runnable, Runnable, Direction]]:
            for runnable in self._nodes.values():
                yield from _runnable_edges(runnable)

        edges = list(_runnables_edges())
        for src, dest, direction in edges:
            self._graph[direction][src.name].add(dest.name)
            self._nodes[src.name] = src
            self._nodes[dest.name] = dest

    def _dfs(self, start_node: str, direction: Direction) -> set[str]:
        def _dfs_trav(node: str) -> Iterator[str]:
            for name in self._graph[direction][node]:
                yield name
                yield from _dfs_trav(name)

        return set(_dfs_trav(start_node))

    @qik.func.cached_property
    def _upstream(self) -> Edges:
        return {name: self._dfs(name, "up") for name in self.nodes}

    @qik.func.cached_property
    def _downstream(self) -> Edges:
        return {name: self._dfs(name, "down") for name in self.nodes}

    ###
    # Filtering methods and filtered properties
    ###

    def filter_caches(self, caches: list[str]) -> Self:
        """Filter the graph by the cache."""
        caches_set = frozenset(c.lower() for c in caches)
        return self.filter(
            (runnable for runnable in self if runnable.cache.lower() in caches_set),
            neighbors=False,
        )

    def filter_cache_status(self, cache_status: qik.conf.CacheStatus) -> Self:
        """Filter the graph by cache status."""

        def _matches_cache_status(entry: Entry | None) -> bool:
            return bool(entry) if cache_status == "warm" else not bool(entry)

        return self.filter(
            (
                runnable
                for runnable in self
                if (
                    runnable.get_cache_backend().type != "none"
                    and _matches_cache_status(runnable.get_cache_entry(artifacts=False))
                )
            ),
            neighbors=False,
        )

    def filter_since(self, git_sha: str) -> Self:
        """Filter the graph since a git SHA"""
        project_dir = qik.conf.root()
        git_dir = pathlib.Path(
            qik.shell.exec("git rev-parse --show-toplevel", check=True).stdout.strip()
        )
        diff_files = qik.shell.exec(f"git diff --name-only {git_sha} -- .", check=True, lines=True)

        # Remember, names from git diff will include folders not in our
        # project root. Make all names relative to the project root
        changes = [
            qik.dep.Glob(str((git_dir / diff_file).relative_to(project_dir)))
            for diff_file in diff_files
        ]

        return self.filter_changes(changes, strategy="since")

    def filter_tags(self, tags: list[str]) -> Self:
        """Filter the graph by tags."""
        tags_set = frozenset(tags)
        return self.filter(
            (runnable for runnable in self if set(runnable.tags) & tags_set),
            neighbors=False,
        )

    def filter_changes(
        self, deps: Iterable[qik.dep.Dep], strategy: qik.runnable.FilterStrategy
    ) -> Self:
        """Filter the graph by a list of changed dependencies."""
        runnables: dict[str, Runnable] = {}
        globs = "\n".join(change.val for change in deps if isinstance(change, qik.dep.Glob))
        if globs:
            runnables |= {
                runnable.name: runnable
                for runnable in self
                if (regex := runnable.filter_regex(strategy)) and regex.search(globs)
            }

        pydists = {change.val for change in deps if isinstance(change, qik.dep.Pydist)}
        if pydists:
            runnables |= {
                runnable.name: runnable
                for runnable in self
                if set(runnable.deps_collection.pydists) & pydists
            }

        return self.filter(runnables.values())

    def filter(self, runnables: Iterable[Runnable], neighbors: bool = True) -> Self:
        """Return a filtered graph. Include upstream and downstream runnables by default."""
        runnables = list(runnables)
        clone = copy.copy(self)
        clone._view = {runnable.name for runnable in runnables if runnable.name in self._nodes}
        if neighbors:
            clone._view |= {dep for runnable in clone._view for dep in self._upstream[runnable]}
            clone._view |= {dep for runnable in clone._view for dep in self._downstream[runnable]}

        if "nodes" in clone.__dict__:
            del clone.__dict__["nodes"]

        if "upstream" in clone.__dict__:
            del clone.__dict__["upstream"]

        return clone

    @qik.func.cached_property
    def upstream(self) -> Edges:
        return {
            name: {dep for dep in deps if self._view is None or dep in self._view}
            for name, deps in self._upstream.items()
            if self._view is None or name in self._view
        }

    @qik.func.cached_property
    def nodes(self) -> dict[str, Runnable]:
        return {
            name: runnable
            for name, runnable in self._nodes.items()
            if self._view is None or name in self._view
        }

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self) -> Iterator[Runnable]:
        return iter(self.nodes.values())

    def __getitem__(self, name: str) -> Runnable:
        return self.nodes[name]


class Runner:
    def __init__(self, graph: Graph) -> None:
        self.graph = graph
        self.pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=qik.ctx.by_namespace("qik").workers
        )
        qik_ctx = qik.ctx.by_namespace("qik")
        if not qik_ctx.logger:
            self.logger = (
                qik.logger.Stdout()
                if len(graph) == 1 or qik_ctx.workers == 1
                else qik.logger.Progress()
            )
        elif qik_ctx.logger == "progress":
            self.logger = qik.logger.Progress()
        else:
            self.logger = qik.logger.Stdout()

    def exec(self, *, changes: Iterable[qik.dep.Dep] | None = None) -> int:
        """Exec the runner, optionally providing a list of changed dependencies."""
        try:
            orig_graph = self.graph
            self.graph = (
                self.graph.filter_changes(changes, strategy="watch") if changes else self.graph
            )
            results = DAGPool(graph=self.graph).exec()
            self.graph = orig_graph
            return max((result.code for result in results.values() if result), default=0)
        finally:
            qik.func.clear_per_run_cache()

    def watch(self) -> None:
        self.logger.print("Watching for changes...", emoji="eyes", color="cyan")
        qik.watcher.start(self)


def _get_graph() -> Graph:
    """Get a filtered graph."""
    qik_ctx = qik.ctx.by_namespace("qik")
    cmds = qik_ctx.commands

    if not cmds:
        cmds = list(qik.cmd.ls())

    try:
        graph = Graph(cmds, modules=qik_ctx.modules, spaces=qik_ctx.spaces)
    except RecursionError as exc:
        raise qik.errors.GraphCycle("Cycle detected in DAG.") from exc

    if qik_ctx.tags:
        graph = graph.filter_tags(qik_ctx.tags)

    if qik_ctx.caches:
        graph = graph.filter_caches(qik_ctx.caches)

    if qik_ctx.cache_status:
        graph = graph.filter_cache_status(qik_ctx.cache_status)

    if qik_ctx.since:
        graph = graph.filter_since(qik_ctx.since)

    return graph


def exec() -> Graph:
    """Run commands based on the current qik context."""
    try:
        qik_ctx = qik.ctx.by_namespace("qik")
        graph = _get_graph()

        if not qik_ctx.ls and not qik_ctx.fail:
            runner = Runner(graph=graph)
            with qik.ctx.set_runner(runner):
                exit_code = runner.exec()

                if qik_ctx.watch:
                    runner.watch()
                else:
                    sys.exit(exit_code)

        return graph
    except Exception as exc:
        qik.errors.print(exc, prefix="Runtime error - ")
        sys.exit(1)



from __future__ import annotations

import pkgutil
from typing import TYPE_CHECKING, Iterator

import msgspec

import qik.conf
import qik.errors
import qik.func
import qik.unset

if TYPE_CHECKING:
    from qik.runnable import Runnable


class Cmd(msgspec.Struct, frozen=True):
    name: str
    runnables: dict[str, Runnable]


@qik.func.cache
def load(name: str, **args: str) -> Cmd:
    """Load a command object."""
    cmd_conf = qik.conf.command(name)
    runnables = pkgutil.resolve_name(cmd_conf.factory or "qik.runnable.factory")(
        name, cmd_conf, **args
    )
    return Cmd(name=name, runnables=runnables)


def runnables(name: str) -> Iterator[Runnable]:
    """Load runnables for a command."""
    try:
        yield from load(name).runnables.values()
    except qik.errors.ArgNotSupplied:

        def _runnable_edges(runnable: Runnable) -> Iterator[Runnable]:
            if runnable.name.startswith(name):
                yield runnable

            for dep in runnable.deps_collection.runnables.values():
                yield from _runnable_edges(dep.obj)

        # Some commands, such as uv.install, take arguments and are included by
        # other commands. Allow the user to type the command name, which will
        # load the graph and return all runnables.
        for cmd in ls():
            for runnable in runnables(cmd):
                yield from _runnable_edges(runnable)


def ls() -> Iterator[str]:
    """List all non-hidden command names."""
    proj = qik.conf.project()
    for module in [None, *proj.modules_by_name, *proj.plugins_by_name]:
        conf = qik.conf.search(module)
        for command in conf.commands:
            cmd_name = f"{module}/{command}" if module else command
            if not qik.conf.command(cmd_name).hidden:
                yield cmd_name



from __future__ import annotations

import fnmatch
import pkgutil
import re
from typing import TYPE_CHECKING, Literal, TypeAlias

import msgspec
from typing_extensions import Self

import qik.cache
import qik.conf
import qik.ctx
import qik.dep
import qik.errors
import qik.file
import qik.func
import qik.hash
import qik.shell
import qik.space
import qik.unset
import qik.venv

if TYPE_CHECKING:
    Direction: TypeAlias = Literal["up", "down"]
    Edges: TypeAlias = dict[str, set[str]]
    FilterStrategy: TypeAlias = Literal["since", "watch"]

    import pathlib

    import qik.cache
    import qik.logger
    import qik.venv


class DepsCollection:
    """A filterable and hashable collection of dependencies for a runnable."""

    def __init__(self, *deps: str | pathlib.Path | qik.dep.Dep, runnable: Runnable):
        self._deps = [
            dep if isinstance(dep, qik.dep.Dep) else qik.dep.Glob(str(dep)) for dep in deps
        ]
        self.runnable = runnable
        self.module = runnable.module
        self.space_globs = runnable.resolved_space.glob_deps if runnable.resolved_space else set()
        self.venv = runnable.resolved_venv

    @property
    def globs(self) -> set[str]:
        return (
            {glob for dep in self._deps for glob in dep.globs}
            | {
                artifact
                for runnable in self.runnables.values()
                for artifact in runnable.obj.artifacts
            }
            | self.venv.glob_deps
            | self.space_globs
        )

    @qik.func.cached_property
    def consts(self) -> set[str]:
        return {
            dep.val for dep in self._deps if isinstance(dep, qik.dep.Const)
        } | self.venv.const_deps

    @qik.func.cached_property
    def watch(self) -> set[str]:
        return {glob for dep in self._deps for glob in dep.watch}

    @qik.func.cached_property
    def since(self) -> set[str]:
        venv_since = self.venv.since_deps if self.runnable.space else set()
        return {glob for dep in self._deps for glob in dep.since} | venv_since

    @property
    def vals(self) -> set[str]:
        return {val for dep in self._deps for val in dep.vals}

    @property
    def pydists(self) -> set[str]:
        return {pydist for dep in self._deps for pydist in dep.pydists}

    @property
    def runnables(self) -> dict[str, qik.dep.Runnable]:
        return {
            runnable.name: runnable
            for dep in self._deps
            for runnable in dep.runnables
            if not self.module or not runnable.obj.module or self.module == runnable.obj.module
        } | self.venv.runnable_deps

    @qik.func.cached_property
    def consts_hash(self) -> str:
        """Hash all consts."""
        return qik.hash.strs(*self.consts)

    def hash_vals(self) -> str:
        """Hash file values."""
        return qik.hash.strs(*self.vals)

    def hash_pydists(self) -> str:
        """Hash python distributions."""
        return qik.hash.pydists(*self.pydists, venv=self.venv)

    def hash_globs(self) -> str:
        """Hash glob pattern."""
        return qik.hash.globs(*self.globs)

    def hash(self) -> str:
        """The full hash."""
        return qik.hash.strs(
            self.consts_hash, self.hash_vals(), self.hash_globs(), self.hash_pydists()
        )


@qik.func.cache
def _num_spaces() -> int:
    return len((set(qik.conf.project().resolved_spaces) | {"default"}))


def fmt_name(
    cmd: str,
    *,
    module: qik.conf.ModuleLocator | None = None,
    space: str | None = "default",
    **args: str,
) -> str:
    module_str = f"#{module.name}" if module else ""
    space_str = f"@{space}" if space and _num_spaces() > 1 else ""
    args_str = "?" + "&".join(f"{k}={v}" for k, v in args.items()) if args else ""
    return f"{cmd}{space_str}{module_str}{args_str}"


def _make_runnable(
    *,
    cmd: str,
    conf: qik.conf.Cmd,
    space: str | None,
    module: qik.conf.ModuleLocator | None = None,
) -> Runnable:
    # If the command has no deps, use "none" as the cache unless explicitly set
    initial_cache = "none" if not conf.deps else conf.cache

    return Runnable(
        name=fmt_name(cmd, module=module, space=space),
        cmd=cmd,
        val=qik.ctx.format(conf.exec, module=module),
        deps=[
            *qik.dep.base(),
            *(qik.dep.factory(dep, module=module, space=space) for dep in conf.deps),
        ],
        tags=conf.tags,
        module=module.name if module else None,
        artifacts=[qik.ctx.format(artifact) for artifact in conf.artifacts],
        cache=qik.ctx.format(
            qik.unset.coalesce(initial_cache, qik.conf.project().defaults.cache, default="local")
        ),
        cache_when=qik.ctx.format(
            qik.unset.coalesce(
                conf.cache_when, qik.conf.project().defaults.cache_when, default="success"
            )
        ),  # type: ignore
        space=space,
    )


def factory(cmd: str, conf: qik.conf.Cmd, **args: str) -> dict[str, Runnable]:
    """The main factory for creating runnables.

    We don't preserve **args in the made runnables because generic runnables do not
    yet support args. Only custom runnables do.
    """
    if "{module" in conf.exec:
        proj = qik.conf.project()
        if not isinstance(conf.space, qik.unset.UnsetType):
            spaces = {conf.space: qik.space.load(conf.space).conf}
        else:
            spaces = proj.resolved_spaces

        runnables = (
            _make_runnable(cmd=cmd, conf=conf, module=module, space=space)
            for space, space_conf in spaces.items()
            for module in space_conf.modules_by_name.values()
        )
    else:
        space = conf.space if not isinstance(conf.space, qik.unset.UnsetType) else "default"
        runnables = [_make_runnable(cmd=cmd, conf=conf, space=space)]

    return {runnable.name: runnable for runnable in runnables}


class Result(msgspec.Struct, frozen=True):
    log: str | None
    code: int
    hash: str

    @classmethod
    def from_cache(cls, entry: qik.cache.Entry) -> Self:
        return cls(log=entry.log, code=entry.manifest.code, hash=entry.manifest.hash)


@qik.func.cache
def _glob_to_regex(glob_pattern: str) -> str:
    """Translate a glob to a regex pattern"""
    return fnmatch.translate(glob_pattern).replace("?s:", "^").replace(r"\Z", "$")


class Runnable(msgspec.Struct, frozen=True, dict=True):
    name: str
    cmd: str
    val: str
    cache: str
    cache_when: qik.conf.CacheWhen
    shell: bool = True
    tags: list[str] = []
    deps: list[qik.dep.Dep] = []
    artifacts: list[str] = []
    module: str | None = None
    args: dict[str, str] = {}
    space: str | None = None
    venv: qik.unset.UnsetType | None = qik.unset.UNSET
    environ: dict[str, str] = {}

    @qik.func.per_run_cached_property
    def description(self) -> str:
        args = ", ".join(f"{k}={v}" for k, v in self.args.items())
        return f"{self.val}({args})" if args and not self.shell else self.val

    @qik.func.cached_property
    def slug(self) -> str:
        return re.sub(r"[^a-zA-Z0-9]", "__", self.name)

    @qik.func.cached_property
    def deps_collection(self) -> DepsCollection:
        return DepsCollection(*self.deps, runnable=self)

    @qik.func.cached_property
    def resolved_space(self) -> qik.space.Space | None:
        return qik.space.load(self.space) if self.space else None

    @qik.func.cached_property
    def resolved_venv(self) -> qik.venv.Venv:
        if not self.resolved_space or self.venv is None:
            return qik.venv.active()
        else:
            return self.resolved_space.venv

    @qik.func.cached_property
    def resolved_environ(self) -> dict[str, str]:
        base_environ = (
            self.resolved_venv.environ if not self.resolved_space else self.resolved_space.environ
        )
        return {**base_environ, **self.environ}

    @property
    def _exec_env(self) -> dict[str, str]:
        """Get the environment for runnables."""
        return {
            **self.resolved_environ,
            "QIK__CMD": self.cmd,
            "QIK__RUNNABLE": self.name,
            "QIK__WORKER": str(qik.ctx.worker_id()),
        }

    def filter_regex(self, strategy: FilterStrategy) -> re.Pattern | None:
        """Generate the regex used for file-based filtering.

        File-based filtering occurs during --since and --watch.
        """
        globs = self.deps_collection.since if strategy == "since" else self.deps_collection.watch
        files_regex = ")|(".join(_glob_to_regex(glob) for glob in globs)
        return re.compile(f"({files_regex})", re.M) if files_regex else None

    @qik.func.cached_property
    def spec_hash(self) -> str:
        """Hash the runnable spec."""
        # Note - we don't include caching arguments in the definition hash since they don't materially
        # affect the outcome.
        return qik.hash.val(
            msgspec.json.encode(msgspec.structs.replace(self, cache="", cache_when=""))
        )

    def hash(self) -> str:
        """Compute the hash, including the command definitions and deps."""
        return qik.hash.strs(self.spec_hash, self.deps_collection.hash())

    def should_cache(self, code: int) -> bool:
        match self.cache_when:
            case "success":
                return code == 0
            case "failed":
                return code != 0
            case "finished":
                return True
            case other:
                raise AssertionError(f'Unexpected cache_when "{other}".')

    def get_cache_backend(self) -> qik.cache.Cache:
        return qik.cache.load(self.cache)

    def get_cache_entry(self, artifacts: bool = True) -> qik.cache.Entry | None:
        if not qik.ctx.by_namespace("qik").force:
            entry = self.get_cache_backend().get(self, artifacts=artifacts)
            if entry and self.should_cache(entry.manifest.code):
                return entry

    def cache_result(self, result: Result) -> None:
        if self.should_cache(result.code):
            self.get_cache_backend().set(self, result)

    def store_deps(
        self,
        path: pathlib.Path,
        *,
        globs: list[str] | None = None,
        pydists: list[str] | None = None,
        hash: bool = True,
    ) -> None:
        """Store serialized dependencies for a runnable."""
        if hash:
            hash_val = DepsCollection(
                *[*(globs or []), *[qik.dep.Pydist(val=pydist) for pydist in pydists or []]],
                runnable=self,
            ).hash()
        else:
            hash_val = None

        qik.file.write(
            path,
            msgspec.json.encode(
                qik.dep.Serialized(globs=globs or [], pydists=pydists or [], hash=hash_val)
            ),
        )

    def _uncached_exec(self, logger: qik.logger.Logger) -> Result:
        """Run a command without wrapped caching."""
        try:
            if self.shell:
                process = qik.shell.popen(self.val, env=self._exec_env)
                output = []
                for line in process.stdout if process.stdout is not None else []:
                    logger.print(line, runnable=self, event="output")
                    output.append(line)

                process.wait()
                code = process.returncode
                log = "".join(output)
            else:
                code, log = pkgutil.resolve_name(self.val)(runnable=self)
                logger.print(log, runnable=self, event="output")
        except qik.errors.RunnableError as exc:
            log = qik.errors.fmt_msg(exc)
            code = 1
            logger.print(log, runnable=self, event="output")

        return Result(log=log, code=code, hash=self.hash())

    def _exec(self) -> Result:
        """Run a command, caching the results."""
        logger = qik.ctx.runner().logger
        print_kwargs = {"runnable": self}

        def _log_start(cached: bool) -> None:
            if cached:
                logger.print(
                    msg=f"{self.name} [default][dim]{self.description}",
                    emoji="fast-forward_button",
                    color="cyan",
                    event="start",
                    **print_kwargs,  # type: ignore
                )
            else:
                logger.print(
                    msg=f"{self.name} [default][dim]{self.description}",
                    emoji="construction",
                    color="cyan",
                    event="start",
                    **print_kwargs,  # type: ignore
                )

        try:
            cache_entry = self.get_cache_entry()
        except qik.errors.RunnableError as exc:
            _log_start(cached=False)
            log = qik.errors.fmt_msg(exc)
            logger.print(f"{log}", runnable=self, event="output")
            result = Result(log=log, code=1, hash="")
        else:
            print_kwargs |= {"cache_entry": cache_entry}

            if cache_entry:
                # Run cached command
                _log_start(cached=True)
                logger.print(cache_entry.log or "", event="output", **print_kwargs)  # type: ignore
                result = Result.from_cache(cache_entry)
            else:
                # Run uncached command
                _log_start(cached=False)
                result = self._uncached_exec(logger=logger)
                self.cache_result(result)

        if result.code == 0:
            logger.print(
                msg=f"{self.name} [default][dim]{self.description}",
                emoji="white_check_mark",
                color="green",
                event="finish",
                result=result,
                **print_kwargs,  # type: ignore
            )
        else:
            logger.print(
                msg=f"{self.name} [default][dim]{self.description}",
                emoji="broken_heart",
                color="red",
                event="finish",
                result=result,
                **print_kwargs,  # type: ignore
            )

        return result

    def exec(self) -> Result:
        with qik.ctx.set_runnable(self), qik.ctx.set_worker_id():
            return self._exec()



from __future__ import annotations

import collections
import functools
import importlib.metadata
import inspect
import os
import pathlib
import re
import sysconfig
import threading
from typing import Iterator

import msgspec

import qik.cmd
import qik.conf
import qik.dep
import qik.errors
import qik.file
import qik.func
import qik.hash


# From jaraco.functools 3.3
def _pass_none(func):
    @functools.wraps(func)
    def wrapper(param, *args, **kwargs):
        if param is not None:
            return func(param, *args, **kwargs)

    return wrapper


# copied from more_itertools 8.8
def _always_iterable(obj, base_type=(str, bytes)):
    if obj is None:
        return iter(())

    if (base_type is not None) and isinstance(obj, base_type):
        return iter((obj,))

    try:
        return iter(obj)
    except TypeError:
        return iter((obj,))


def _top_level_declared(dist):
    return (dist.read_text("top_level.txt") or "").split()


def _top_level_inferred(dist):
    opt_names = {
        f.parts[0] if len(f.parts) > 1 else inspect.getmodulename(f)
        for f in _always_iterable(dist.files)
    }

    @_pass_none
    def importable_name(name):
        return "." not in name

    return filter(importable_name, opt_names)


@qik.func.cache
def _normalize_pydist_name(name: str) -> str:
    return re.sub(r"[-_.]+", "-", name).lower().replace("-", "_")


@qik.func.cache
def _pydist_version_overrides() -> dict[str, str]:
    project_conf = qik.conf.project()
    base = collections.defaultdict(str) if project_conf.pydist.ignore_missing else {}
    return base | {
        _normalize_pydist_name(name): version
        for name, version in project_conf.pydist.versions.items()
    }


class PackagesDistributions(msgspec.Struct):
    venv_hash: str
    packages_distributions: dict[str, list[str]]


class Venv(msgspec.Struct, frozen=True, dict=True):
    name: str
    conf: qik.conf.Venv

    def __post_init__(self):
        self.__dict__["_packages_distributions_lock"] = threading.Lock()
        self.__dict__["_packages_distributions"] = (None, None)

    # TODO: Consider caching distributions and all versions in a similar manner as
    # the packages_distributions cache.
    def distributions(self, **kwargs) -> Iterator[importlib.metadata.Distribution]:
        return importlib.metadata.distributions(path=[str(self.site_packages_dir)], **kwargs)  # type: ignore

    @property
    def alias(self) -> str:
        return f'"{self.name}" space'

    def version(self, name: str) -> str:
        try:
            return next(self.distributions(name=name)).version
        except StopIteration:
            try:
                return _pydist_version_overrides()[_normalize_pydist_name(name)]
            except KeyError as exc:
                raise qik.errors.DistributionNotFound(
                    f'Distribution "{name}" not found in {self.alias}.'
                ) from exc

    def packages_distributions(self) -> dict[str, list[str]]:
        """Obtain a mapping of modules to their associated python distributions.

        This is an expensive command, so use an underlying cache when possible.
        """
        venv_contents = set(os.listdir(self.site_packages_dir)) - {"__pycache__"}
        venv_hash = qik.hash.strs(*sorted(venv_contents))
        with self.__dict__["_packages_distributions_lock"]:
            if self.__dict__["_packages_distributions"][0] != venv_hash:
                pydist_conf = qik.conf.project().pydist
                cache_path = (
                    qik.conf.priv_work_dir()
                    / "venv"
                    / ".packages_distributions"
                    / f"{self.name}.json"
                )
                overrides = (
                    {}
                    if not pydist_conf.ignore_missing_modules
                    else collections.defaultdict(lambda: [""])
                )
                overrides |= {module: [dist] for module, dist in pydist_conf.modules.items()}
                try:
                    cached_val = msgspec.json.decode(
                        cache_path.read_bytes(), type=PackagesDistributions
                    )
                    if cached_val.venv_hash == venv_hash:
                        return overrides | cached_val.packages_distributions
                except FileNotFoundError:
                    pass

                pkg_to_dist = collections.defaultdict(list)
                for dist in self.distributions():
                    for pkg in _top_level_declared(dist) or _top_level_inferred(dist):
                        pkg_to_dist[pkg].append(dist.metadata["Name"])

                cached_val = PackagesDistributions(
                    venv_hash=venv_hash,
                    packages_distributions=dict(pkg_to_dist),
                )
                qik.file.write(cache_path, msgspec.json.encode(cached_val))

                self.__dict__["_packages_distributions"] = (
                    venv_hash,
                    overrides | cached_val.packages_distributions,
                )

            return self.__dict__["_packages_distributions"][1]

    @qik.func.cached_property
    def reqs(self) -> list[str]:
        return self.conf.reqs if isinstance(self.conf.reqs, list) else [self.conf.reqs]

    @property
    def environ(self) -> dict[str, str]:
        return os.environ  # type: ignore

    @property
    def dir(self) -> pathlib.Path:
        raise NotImplementedError

    @property
    def site_packages_dir(self) -> pathlib.Path:
        raise NotImplementedError

    @qik.func.cached_property
    def lock(self) -> str | None:
        return self.conf.lock

    @qik.func.cached_property
    def runnable_deps(self) -> dict[str, qik.dep.Runnable]:
        raise NotImplementedError

    @qik.func.cached_property
    def glob_deps(self) -> set[str]:
        return {self.lock} if self.lock else set(self.reqs)

    @qik.func.cached_property
    def const_deps(self) -> set[str]:
        """Return the serialized venv as a constant dep."""
        return {msgspec.json.encode(self).decode()}

    @property
    def since_deps(self) -> set[str]:
        return set().union(self.reqs, [self.lock] if self.lock else [])


class Active(Venv, frozen=True, dict=True):
    """
    The active virtual environment.
    """

    conf: qik.conf.ActiveVenv

    @property
    def alias(self) -> str:
        return "active virtual environment"

    @property
    def dir(self) -> pathlib.Path:
        return pathlib.Path(sysconfig.get_path("data"))

    @property
    def site_packages_dir(self) -> pathlib.Path:
        return pathlib.Path(sysconfig.get_path("purelib"))

    @qik.func.cached_property
    def runnable_deps(self) -> dict[str, qik.dep.Runnable]:
        return {}


_ACTIVE: Active = Active(name=".active", conf=qik.conf.ActiveVenv())


def active() -> Active:
    """Return the active venv."""
    return _ACTIVE



from __future__ import annotations

import pathlib
import pkgutil
import re
from typing import TYPE_CHECKING

import msgspec

import qik.cmd
import qik.conf
import qik.ctx
import qik.errors
import qik.file
import qik.func
import qik.hash
import qik.unset

if TYPE_CHECKING:
    import qik.runnable


@qik.func.cache
def _normalize_pydist_name(pydist: str) -> str:
    pydist = pydist.lower().strip()
    return re.sub(r"[^a-z0-9]+", "-", pydist)


class Runnable(msgspec.Struct, frozen=True):
    """A runnable included by a dependency."""

    name: str
    obj: qik.runnable.Runnable
    strict: bool = False
    isolated: bool | qik.unset.UnsetType = qik.unset.UNSET


class Dep(msgspec.Struct, frozen=True, tag=True, dict=True):
    """A base dependency."""

    val: str

    def __str__(self) -> str:
        return self.val

    @qik.func.cached_property
    def runnables(self) -> list[Runnable]:
        """Return any runnables of the dependency."""
        return []

    @qik.func.cached_property
    def globs(self) -> list[str]:
        """Return any glob patterns of the dependency."""
        return []

    @qik.func.cached_property
    def pydists(self) -> list[str]:
        """Return any pydists of the dependency."""
        return []

    @qik.func.cached_property
    def vals(self) -> list[str]:
        """Return any vals of the dependency."""
        return []

    @qik.func.cached_property
    def watch(self) -> list[str]:
        """Return any globs used for --watch option.

        When using --watch, we need to react to events based on file
        system changes. Similar to `since`, we return globs
        used by the watcher to determine changes. The main difference
        between `since` and `watch` is that the watcher has access to
        more than just git files (venv, etc). As a result, we don't
        need to react to as many files in the git repo, such as
        reacting based on a lock file change.
        """
        return self.globs

    @qik.func.cached_property
    def since(self) -> list[str]:
        """Return any globs used for --since option.

        When using --since, we need to estimate which commands need to
        be re-executed based on git changes. For deps such as
        pydists or vals that aren't directly tied to git, we return globs
        that encapsulate those changes (lock files, parent files, etc).
        """
        return self.watch


def factory(
    conf: qik.conf.Dep | str | pathlib.Path,
    module: qik.conf.ModuleLocator | None = None,
    space: str | None = None,
) -> Dep:
    """A factory for creating dependencies from a configuration."""

    def _fmt(val: str) -> str:
        return qik.ctx.format(val, module=module)

    match conf:
        case str() | pathlib.Path():
            return Glob(_fmt(str(conf)))
        case qik.conf.GlobDep():
            return Glob(_fmt(conf.pattern))
        case qik.conf.PydistDep():
            return Pydist(_fmt(conf.name))
        case qik.conf.ValDep():
            return Val(_fmt(conf.key), file=_fmt(conf.file))
        case qik.conf.CmdDep():
            return Cmd(_fmt(conf.name), strict=conf.strict, isolated=conf.isolated)
        case qik.conf.ConstDep():
            return Const(_fmt(conf.val))
        case qik.conf.LoadDep():
            return Load(_fmt(conf.path), default=conf.default)
        case other:
            factory = qik.conf.get_type_factory(other)
            return pkgutil.resolve_name(factory)(conf, module=module, space=space)


class Glob(Dep, frozen=True):
    """A dependent glob pattern"""

    @qik.func.cached_property
    def globs(self) -> list[str]:
        return [self.val]


class Val(Dep, frozen=True):
    """A value from a file."""

    file: str

    @property
    def vals(self) -> list[str]:  # type: ignore
        # This dependency is still experimental and only used by qik.pygraph.
        raise NotImplementedError

    @qik.func.cached_property
    def watch(self) -> list[str]:
        return [self.file]


class BaseCmd(Dep, frozen=True):
    strict: bool = False
    isolated: bool | qik.unset.UnsetType = qik.unset.UNSET
    args: dict[str, str | None] = {}

    def get_cmd_name(self) -> str:
        raise NotImplementedError

    def get_cmd_args(self) -> dict[str, str | None]:
        return self.args

    @qik.func.cached_property
    def runnables(self) -> list[Runnable]:
        """Return any runnables of the dependency."""
        return [
            Runnable(name=runnable.name, obj=runnable, strict=self.strict, isolated=self.isolated)
            for runnable in qik.cmd.load(
                self.get_cmd_name(), **self.get_cmd_args()
            ).runnables.values()
        ]


class Cmd(BaseCmd, frozen=True):
    """A dependent command."""

    def get_cmd_name(self) -> str:
        return self.val


class Pydist(Dep, frozen=True):
    """A python distribution dependency."""

    @property
    def normalized(self) -> str:
        return _normalize_pydist_name(self.val)

    @qik.func.cached_property
    def pydists(self) -> list[str]:
        return [self.val]


class Const(Dep, frozen=True):
    """A constant value."""

    @qik.func.cached_property
    def since(self) -> list[str]:
        return ["*qik.toml"]


class Load(Dep, frozen=True):
    """Load serialized dependencies from a file."""

    default: list[str] = []  # The default globs if the file doesn't exist

    def load(self) -> Serialized | None:
        """Get the loaded dependencies."""
        try:
            return msgspec.json.decode(pathlib.Path(self.val).read_bytes(), type=Serialized)
        except (FileNotFoundError, msgspec.DecodeError):
            return None

    @property
    def pydists(self) -> list[str]:  # type: ignore
        deps = self.load()
        return deps.pydists if deps else []

    @property
    def globs(self) -> list[str]:  # type: ignore
        deps = self.load()
        return deps.globs if deps else self.default


class Serialized(msgspec.Struct, frozen=True, omit_defaults=True):
    """A serialized representation of dependencies, meant to be loaded from a file."""

    globs: list[str] = []
    pydists: list[str] = []
    hash: str | None = None


@qik.func.cache
def base() -> list[Dep]:
    """The base dependencies for the project."""
    return [factory(dep) for dep in qik.conf.project().base.deps]
